#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "cell.h"
#include "map.h"
#include "defs.h"
#include "v2.h"
#include "v1.h"

#define BITS_V1_CELL	38
#define BITS_V2_CELL	57
#define VERSION_SIZE	4
#define CONFIG_FLAGS	"0000"
#define DIM_SIZE	8
static int allocate_config_memory(char** target, int dim[], int cell_size);
static int calc_config_size(int dim[]);
static int write_ascii_config(char* config, int dim[], FILE* target, 
				int cell_size, int version);
static void write_header(FILE* target, int dim[],int version);
static void write_version(FILE* target, int version );
static void write_dim(FILE* target, int dim);
static int bits_per_cell(int version);


int generate_result(struct cell_desc** map,int dim[],FILE* target, int version) {
	int i,j;
	struct v2_part* v2p = NULL;
	char* config = NULL;
	int pos = 0;
	int cell_size = bits_per_cell(version);
	
	if ( allocate_config_memory(&config,dim,cell_size) ) {
		return 1;
	}
	
	for ( i = 0 ; i < dim[1] ; i++ ){
		for ( j = 0 ; j < dim[0] ; j++ ) {
			if ( MAP(map,dim,j,i) != NULL ) {
				//generate_v1_part(&v1p,MAP(map,dim,j,i)->list);
				generate_v2_part(&v2p,MAP(map,dim,j,i)->list,version);
				print_v2_part(v2p,version);
				printf("STREAM: ");
				serialize_v2_part(v2p,config+pos,version);
				destroy_v2_part(&v2p);
				printf("\n");
			}
			pos += cell_size;
		}
	}
	write_ascii_config(config,dim,target,cell_size,version);
	free(config);
	return 0;
}

static int allocate_config_memory(char** target, int dim[], int cell_size) {
	int size = calc_config_size(dim)*cell_size;
	*target = malloc(size);
	
	if ( *target == NULL ) {
		fprintf(stderr,"Memory allocation error in backend\n");
		return 1;
	}
	
	memset(*target,'0',size);
	
	return 0;
}

static int calc_config_size(int dim[]) {
	int cells = 1;
	int i = 0;
	for ( i = 0 ; i < MAX_DIMS ; i++ ) {
		cells = cells * dim[i];
	}
	return cells;
}

static int write_ascii_config(char* config, int dim[], FILE* target,
				int cell_size,int version) {
	int bytes = calc_config_size(dim)*cell_size;
	int i = 0;
	
	if ( target == NULL) {
		fprintf(stderr, "NULL filedescriptor\n");
		return 1;
	}
	
	write_header(target,dim,version-1);
	
	for ( i = 0 ; i < bytes; i++ ) {
		if ( i > 0 && (i % cell_size) == 0 ) {
			fprintf(target,"\n");
		}
		fprintf(target,"%c",*(config+i));
	}
}

static void write_header(FILE* target, int dim[],int version) {
	int i = 0 ;
	write_version(target,version);
	fprintf(target,"%s",CONFIG_FLAGS);
	for ( i = 0 ; i < MAX_DIMS ; i++ ) {
		//fprintf(target," %d", dim[i]);
		write_dim(target,dim[i]);
	}
	fprintf(target,"\n");
}

static void write_version(FILE* target, int version ) {
	int i = VERSION_SIZE -1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(target,"%c",(char)(((version>>i)&0x01)+'0'));
	}
}

static void write_dim(FILE* target, int dim) {
	int i = DIM_SIZE -1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(target,"%c",(char)(((dim>>i)&0x01)+'0'));
	}
}

static int bits_per_cell(int version ) {

	switch(version) {
	case 1:
		return BITS_V1_CELL;
		break;
	case 2:
		return BITS_V2_CELL;
		break;
	default:
		fprintf(stderr,"Invalid version size\n");
		return 0;
	}
}
