#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"

void free_statements(struct statements* first);
void free_cell(struct cell_desc* cell);


const char* statement_names[] = {
	"INIT",
	"ADD",
	"SUB",
	"MUL",
	"DIV",
	"SLR",
	"SLL",
	"AND",
	"OR",
	"NOR",
	"XOR",
	"MOV",
	"BRANCH",
	"SWITCH",
	"SELECT",
	"VALIDWRITE",
	"WRITEPOLICY"
};

const char* operand_names[] = {
	"UNKNOWN",
	"NORTH",
	"WEST",
	"EAST",
	"SOUTH",
	"STATE",
	"ALUOUT",
	"MULOUT"
};

const char* flag_names[] = {
	"NONE",
	"ZERO",
	"NEGATIVE",
	"VALID",
	"NEVER"
};

const char* invert_names[] = {
	"INVERT",
	"NO INVERT"
};

const char* write_flag_names[] = {
	"ALWAYS",
	"VALID"
};

struct cell_desc* first_cell = NULL;

void print_cells() {
	struct cell_desc* next = first_cell;
	struct statements* s = NULL;
	while ( next != NULL ) {
		printf("Cell: %d %d\n",next->pos[0],next->pos[1]);
		s = next->list;
		print_statement_list(s);
		
		next = next->next;
	}
}

void print_statement_list(struct statements* list) {
	struct statements* s = list;
	while ( s != NULL ) {
		print_statement(stdout,s);
		if ( s->children != NULL ) {
			print_statement_list(s->children);
		}
		
		s = s->next;
	}
}

void print_statement(FILE* fd, struct statements* s) {
	if ( s == NULL || fd == NULL) {
		return;
	}
	fprintf(fd,"Type: %s\n",str_type(s->stype));
	fprintf(fd,"Flags: %s\n",str_flag(s->flag));
	fprintf(fd,"Operands: %s %s\n",operand_names[s->op1], operand_names[s->op2]);
	fprintf(fd,"Target: %s\n",operand_names[s->target]);
	if ( s->stype == s_writepolicy ) {
		fprintf(fd,"Write flags: %s\n",str_wr_flag(s->wr_flag));
	}
	if ( s->stype == s_branch && s->flag == br_valid ) {
		fprintf(fd,"Data: %s\n",str_invert(s->data));
	} else {
		fprintf(fd,"Data: %d\n",s->data);
	}
	fprintf(fd,"Version: %d\n\n",s->version);
}

struct statements* new_statement() {
	struct statements* stmnt = NULL;
	stmnt = malloc(sizeof(struct statements));
	if ( stmnt == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		exit(-1);
	}
	memset(stmnt,0,sizeof(struct statements));
	return stmnt;
}

void push_cell() {
	struct cell_desc* cell = NULL;
	cell = malloc(sizeof(struct cell_desc));
	if ( cell == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		exit(-1);
	}
	memset(cell,0,sizeof(struct cell_desc));
	cell->next = first_cell;
	first_cell = cell;
}

void pop_cell() {
	struct cell_desc* remove = first_cell;
	if ( first_cell != NULL ) {
		first_cell = first_cell->next;
	}
	free_cell(remove);
}

void free_cell_list(struct cell_desc* list) {
	struct cell_desc* current = list;
	struct cell_desc* next = NULL;
	
	while ( current != NULL ) {
		next = current->next;
		free_cell(current);
		current = next;
	}
	
}

void free_cell(struct cell_desc* cell) {
	free_statements(cell->list);
	free(cell);

}

void free_statements(struct statements* first) {
	struct statements* next = first;
	struct statements* temp = NULL;
	while ( next != NULL ) {
		free_statements(next->children);
		temp = next->next;
		free(next);
		next = temp;
	}
}

void print_cell_pos(FILE* fd,struct cell_desc* cell) {
	int i = 0;
	for ( ; i < MAX_DIMS ; i++ ) {
		fprintf(fd,"[%d]",cell->pos[i]);
	}
}

const char* str_type(enum statement_type s) {
	return statement_names[s];
}

const char* str_op(enum operands op) {
	return operand_names[op];
}

const char* str_flag(enum branch_flags flag) {
	return flag_names[flag];
}

const char* str_invert(enum invert inv) {
	return invert_names[inv];
}

const char* str_wr_flag(enum write_flags wr) {
	return write_flag_names[wr];
}
