#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "cell.h"
#include "map.h"

static int allocate_grid(struct cell_desc*** array, int dim[]);
static int fill_dummies(struct cell_desc*** map, int dim[]);
static void build_map(struct cell_desc* list, int dim[], 
			struct cell_desc*** map);

int build_grid(struct cell_desc* first, int dim[], struct cell_desc*** map){
	if ( MAX_DIMS != 2 ) {
		fprintf(stderr,"Grid builder is 2D only for now\n");
		return 1;
	}
	
	if ( allocate_grid(map,dim) ) {
		fprintf(stderr,"Memory allocation error\n");
		return 1;
	}
	
	build_map(first,dim,map);
	return 0;
}

static int allocate_grid(struct cell_desc*** array, int dim[]) {
	*array = malloc(sizeof(struct cell_desc*)*dim[0]*dim[1]);
	
	if ( *array == NULL ) {
		return 1;
	}
	
	memset(*array,0,sizeof(struct cell_desc*)*dim[0]*dim[1]);
	
	return 0;
}

static void build_map(struct cell_desc* list, int dim[], 
			struct cell_desc*** map) {
	
	struct cell_desc* current = list;
	
	while ( current != NULL ) {
		MAP(*map,dim,current->pos[0],current->pos[1]) = current;
		current = current->next;
	}
	
}

void print_map(struct cell_desc** map, int dim[]) {
	int i,j;
	
	for ( i = 0 ; i < dim[0] ; i++ ) {
		for ( j = 0 ; j < dim[1] ; j++ ) {
			printf("%p\t",MAP(map,dim,i,j));
		}
		printf("\n");
	}
}

int count_empties(struct cell_desc** map, int dim[]) {
	int i,j;
	int res = 0;
	for ( i = 0 ; i < dim[0] ; i++ ) {
		for ( j = 0 ; j < dim[1] ; j++ ) {
			if ( MAP(map,dim,i,j) == NULL ) {
				res++;
			}
		}
	}
	return res;
}

void free_map(struct cell_desc*** map) {
	free(*map);
	*map = NULL;
}
