#include <stdio.h>
#include <stdint.h>

#include "cell.h"
#include "rc_manager.h"
#include "counters.h"

void initialize_cell(cell_info* cell) {
	cell->config_state = UNCONFIGURED;
	cell->config_bus = NULL;
	cell->ticks_left = 0;
	cell->register_valid_north = 0;
	cell->register_north = 0;
	cell->register_valid_west = 0;
	cell->register_west = 0;
	/* Set some visible value here.. as the init values should 
	 * never be used */
	cell->result_valid_east = 0;
	cell->result_east = 0xFCFC;
	cell->result_valid_south = 0; 
	cell->result_south = 0xFCFC; 

	cell->comp_state.valid_state = 0;
	cell->comp_state.state = 0;
	
	cell->comp_state.fabric = NULL;
}


void cell_tick(cell_info* cell, struct counters* cnt){
	printf("Tick @ cell %p\n",cell);
	
	/* TODO: Reconfigure */
	switch(cell->config_state) {
	case UNCONFIGURED:
		printf("Unconfigured tick\n");
		break;
		
	
	case RECONFIGURING:
		printf("Reconfiguration tick\n");
		cell->result_valid_east = 0;
		cell->result_valid_south = 0;
		if ( cell->ticks_left == 1 ) {
			printf("Reconfigration tick(cell->transfer)\n");
			cell->config_bus->drain = cell->comp_state;
			cell->comp_state = cell->config_bus->source;
		} else {
			printf("Reconfigration tick %d\n", cell->ticks_left);
			cell->ticks_left--;
		}
		record_rc_tick(cnt);
		break;
	
	case COMPUTING:
		printf("Execution tick\n");
		/* Execute */
		if ( cell->comp_state.fabric != NULL ) {
			printf("Cell wants to execute function: %p\n",
				cell->comp_state.fabric
			);
			record_computation_tick(cnt);
			cell->comp_state.fabric(
				cell->ticks_left,
				cell->register_valid_north,
				cell->register_north,
				cell->register_valid_west,
				cell->register_west,
				&(cell->result_valid_east),
				&(cell->result_east),
				&(cell->result_valid_south),
				&(cell->result_south),
				&(cell->comp_state.valid_state),
				&(cell->comp_state.state)
			);
		} else {
			printf("ERROR: Cell in COMPUTING-state without fabric\n");
		}
		
		if ( cell->ticks_left == 1 ) {
			printf("Ticks left reset to %d\n",DATA_TRANSFER_TICKS);
			cell->ticks_left = DATA_TRANSFER_TICKS;
		} else {
			cell->ticks_left--;
		}
		break;
	
	default:
		printf("ERROR: Default reached in state-switch. %d\n",
			cell->config_state);
		break;
	}
}
