#include <stdio.h>

#include "counters.h"
#include "library.h"
#include "memory.h"
#include "rc_manager.h"
#include "cell.h"


static void record_memory_access(struct counters* cnt);
static void record_rc_memory_access(struct counters* cnt);
static void record_memory_write(struct counters* cnt);
static void record_memory_read(struct counters* cnt);
static void record_rc_memory_write(struct counters* cnt);
static void record_rc_memory_read(struct counters* cnt);
static void record_cum_read_size(struct counters* cnt, int amount);
static void record_cum_write_size(struct counters* cnt, int amount);

static void record_memory_access(struct counters* cnt) {
	cnt->total_memory_accesses++;
	cnt->tick_memory_accesses++;
	return;
}

static void record_memory_write(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_memory_writes++;
	record_memory_access(cnt);
}

static void record_memory_read(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_memory_reads++;
	record_memory_access(cnt);
}

void record_memory_write_size(struct counters* cnt,int amount) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_memory_write_size += amount;
	record_cum_write_size(cnt,amount);
	record_memory_write(cnt);
}

void record_memory_read_size(struct counters* cnt,int amount) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_memory_read_size += amount;
	record_cum_read_size(cnt,amount);
	record_memory_read(cnt);
}

static void record_rc_memory_access(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->total_rc_memory_accesses++;
	cnt->tick_rc_memory_accesses++;
	return;
}

static void record_rc_memory_write(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_rc_memory_writes++;
	record_rc_memory_access(cnt);
}

static void record_rc_memory_read(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_rc_memory_reads++;
	record_rc_memory_access(cnt);
}

void record_rc_memory_write_size(struct counters* cnt,int amount) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_rc_memory_write_size += amount;
	record_cum_write_size(cnt,amount);
	record_rc_memory_write(cnt);
}

void record_rc_memory_read_size(struct counters* cnt,int amount) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_rc_memory_read_size += amount;
	record_cum_read_size(cnt,amount);
	record_rc_memory_read(cnt);
}


static void record_cum_write_size(struct counters* cnt, int amount) {
	cnt->tick_cumulative_write_size += amount;
}

static void record_cum_read_size(struct counters* cnt, int amount) {
	cnt->tick_cumulative_read_size += amount;
}

void record_rc_tick(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->total_rc_ticks++;
	cnt->tick_rc_ticks++;
	return;
}

void record_computation_tick(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->total_computation_ticks++;
	cnt->tick_computation_ticks++;
	return;
}

void reset_tick_record(struct counters* cnt) {
	if ( cnt == NULL ) {
		return;
	}
	cnt->tick_memory_accesses = 0;
	cnt->tick_memory_writes = 0;
	cnt->tick_memory_reads = 0;
	cnt->tick_memory_write_size = 0;
	cnt->tick_memory_read_size = 0;
	cnt->tick_rc_memory_accesses = 0;
	cnt->tick_rc_memory_writes = 0;
	cnt->tick_rc_memory_reads = 0;
	cnt->tick_rc_memory_write_size = 0;
	cnt->tick_rc_memory_read_size = 0;
	cnt->tick_cumulative_write_size = 0;
	cnt->tick_cumulative_read_size = 0;
	cnt->tick_rc_ticks = 0;
	cnt->tick_computation_ticks = 0;
	return;
}

int initialize_records(char* stats_file,FILE** fd, int x, int y) {
	if ( stats_file == NULL || fd == NULL ) {
		return 0;
	}
	*fd = fopen(stats_file,"w");
	
	if ( *fd == NULL ) {
		perror("initialize records:");
		return 0;
	}
	fprintf(*fd,"# Gridsize - X: %d Y: %d\n",x,y);
	fprintf(*fd,"# Cells	= %d\n",x*y);
	fprintf(*fd,"# Csize	= %d (%d)\n",
		RC_SIZE*x*y,
		configuration_size(x,y)
	);
	fprintf(*fd,"# Msize	= %d (%d)\n",
		MEMORY_RC_SIZE*2*(x+y),
		memory_configuration_size(x,y)
	);
	fprintf(*fd,"# Dsize	= %d (%d)\n",
		DATA_PORT_SIZE,
		sizeof(cell_transfer)
	);
	fprintf(*fd,"#T\tMEM\tR\tW\tRs\tWs\tRCMEM\tR\tW\tRs\tWs\tCWs\tCRs\tRC\tComp\n"
	);
	return 1;
}

void finalize_records(FILE** fd) {
	if ( fd == NULL ) {
		return;
	}
	fclose(*fd);
	*fd = NULL;
}

void save_tick_record(FILE* fd, int tick, struct counters* cnt) {
	fprintf(fd,"%d\t",tick);
	fprintf(fd,"%lld\t",cnt->tick_memory_accesses);
	fprintf(fd,"%lld\t",cnt->tick_memory_reads);
	fprintf(fd,"%lld\t",cnt->tick_memory_writes);
	fprintf(fd,"%lld\t",cnt->tick_memory_read_size);
	fprintf(fd,"%lld\t",cnt->tick_memory_write_size);
	fprintf(fd,"%lld\t",cnt->tick_rc_memory_accesses);
	fprintf(fd,"%lld\t",cnt->tick_rc_memory_reads);
	fprintf(fd,"%lld\t",cnt->tick_rc_memory_writes);
	fprintf(fd,"%lld\t",cnt->tick_rc_memory_read_size);
	fprintf(fd,"%lld\t",cnt->tick_rc_memory_write_size);
	fprintf(fd,"%lld\t",cnt->tick_cumulative_read_size);
	fprintf(fd,"%lld\t",cnt->tick_cumulative_write_size);
	fprintf(fd,"%lld\t",cnt->tick_rc_ticks);
	fprintf(fd,"%lld\n",cnt->tick_computation_ticks);
}
