#include <stdio.h>
#include <stdlib.h>

#include "grid.h"
#include "cell.h"
#include "counters.h"

static void grid_cell_transfer(struct grid_cell* gcell);
static int grid_create_cell(struct grid_cell** gcell, int x, int y);


/* Create a grid with the given dimentions */
int grid_create(int x, int y, struct grid_cell** NW_corner) {
	struct grid_cell* start = NULL;
	struct grid_cell* west_cell = NULL;
	struct grid_cell* north_cell = NULL;
	struct grid_cell* current_row = NULL;
	struct grid_cell* current = NULL;
	
	current_row = start;
	
	for ( int i = 0 ; i < y ; i++ ) {
		
		/* Create row head */
		if ( !grid_create_cell(&current,0,i) ) {
			fprintf(stderr,"Memory allocation error\n");
			return 0;
		}
		if ( i == 0 ) {
			start = current;
		}
		
		north_cell = current_row;
		current_row = current;
		west_cell = current;
		
		if ( north_cell != NULL ) {
			north_cell->south = current;
			north_cell = north_cell->east;
		}
		
		current = NULL;
		for ( int j = 1 ; j < x ; j++ ) {
			if ( !grid_create_cell(&current,j,i) ) {
				fprintf(stderr,"Memory allocation error\n");
				return 0;
			}
			
			/* Add pointers */
			/* Should never happen */
			if ( west_cell != NULL ) {
				west_cell->east = current;
				west_cell = current;
			} 
			
			if ( north_cell != NULL ) {
				north_cell->south = current;
				north_cell = north_cell->east;
			} 
		}
		west_cell = NULL;
		north_cell = NULL;
	}
	*NW_corner = start;
	return 1;
}

void grid_free(struct grid_cell** NW_corner) {
	struct grid_cell* next_row;
	struct grid_cell* current_row; 
	struct grid_cell* current_col;
	struct grid_cell* next_col;
	
	//next_row = NW_corner->south;
	current_row = *NW_corner;
	while ( current_row != NULL ) {
		next_row = current_row->south;
		current_col = current_row;
		while( current_col != NULL ) {
			next_col = current_col->east;
			free(current_col);
			current_col = next_col;
		}
		current_row = next_row;
	}
	*NW_corner = NULL;
}

static int grid_create_cell(struct grid_cell** gcell, int x, int y) {
	static int cell = 0;
	struct grid_cell* new_grid_cell;
	fprintf(stderr,"Creating cell: %d\n", ++cell);
	
	new_grid_cell = malloc(sizeof(struct grid_cell));
	if ( new_grid_cell == NULL ) {
		return 0;
	} else {
		initialize_cell(&(new_grid_cell->cell));
		new_grid_cell->east = NULL;
		new_grid_cell->south = NULL;
		new_grid_cell->x = x;
		new_grid_cell->y = y;
		*gcell = new_grid_cell;
		return 1;
	}
}

void grid_internal_register_transfer(struct grid_cell* NW_corner) {
	struct grid_cell* current_row = NW_corner;
	struct grid_cell* current_col = NULL;
	
	while( current_row != NULL ) { 
		current_col = current_row;
		while ( current_col != NULL ) {
			grid_cell_transfer(current_col);
			current_col = current_col->east;
		}
		current_row = current_row->south;
	}
	
}

/* Transfer computed values to registers in next cell in grid */
static void grid_cell_transfer(struct grid_cell* gcell) {
	/* TODO: Validate gcell */
	struct grid_cell* east = gcell->east;
	struct grid_cell* south = gcell->south;
	
	
	if ( east != NULL ) {
		east->cell.register_valid_west = gcell->cell.result_valid_east;
		east->cell.register_west = gcell->cell.result_east;
	
	} else {
		printf("%d %d\n",gcell->x,gcell->y);
		printf("Transfering %lld - 0x%llx out of world @ %d,%d\n", 
			gcell->cell.result_east,
			gcell->cell.result_east,
			gcell->x,
			gcell->y
		);
	}

	if ( south != NULL ) {
		south->cell.register_valid_north = gcell->cell.result_valid_south;
		south->cell.register_north = gcell->cell.result_south;
	
	} else {
		printf("Transfering %lld - 0x%llx out of world @ %d,%d\n", 
			gcell->cell.result_south,
			gcell->cell.result_south,
			gcell->x,
			gcell->y
		);
	}
}


void grid_tick(struct grid_cell* NW_corner, struct counters* cnt) {
	struct grid_cell* current_row = NW_corner;
	struct grid_cell* current_col = NULL;
	
	while( current_row != NULL ) { 
		current_col = current_row;
		while ( current_col != NULL ) {
			/* TICK */
			cell_tick(&(current_col->cell),cnt);
			current_col = current_col->east;
		}
		current_row = current_row->south;
	}
}
