#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "grid.h"
#include "print_utils.h"
#include "rc_manager.h"
#include "memory.h"
#include "library.h"
#include "counters.h"

#define SWITCH_RATE(x,y) (((x)+(y)+2)*RC_TICKS+RC_TICKS)
#define SWITCH_IR(x,y) (((x)+(y)+2)+2)
#define ADDRESS_SIZE 8

#define PARAMS_MIN 7
#define RC_OFFSET 4

#define PRE_TRIGGER 6

//#define DEBUG

struct task_info {
	struct libraries* current;
};

static int load_libraries(int argc, char** argv, int x, int y, 
			uint8_t* memory_target,
			struct libraries** list);
static void switch_tasks(int x, int y,
		struct task_info* task_info, 
		struct libraries* list,
		struct rc_manager* rc_list,
		struct mem_interface_rc* mem_units,
		uint8_t* memory_base);
static int parse_arguments(int argc, char** argv, int* x, int* y, 
			char** stats_file, char** io_file,
			int* idle_size, int* config_size);
static void print_rc(FILE* rc_dump, char* ctrl, 
			int source_address, int drain_address,
			int idles);
static void print_address(FILE* target, int address);
static void perform_io_dump(FILE* target, FILE* rc_data,
			struct mem_interface_rc* mem_units,
			int idle_size);
static void fseek_rc(FILE* rc_data);
static void drain_rc(FILE* target, FILE* rc_data);
static int initialize_io_dump(FILE** io_dump, char* file_name);
static int initialize_rc_dump(FILE** rc_dump);


int main (int argc, char** argv) {
	int x,y;
	int global_ticks = -12; /* spinup */
	struct grid_cell* gstart = NULL;
	struct rc_manager* rc_list = NULL;
	struct mem_interface_rc* mem_units = NULL;
	struct libraries* libs = NULL;
	uint8_t* memory_base = NULL;
	struct counters cnt = {0,};
	char* stats_file = NULL;
	FILE* fd;
	char* io_file = NULL;
	FILE* io_dump = NULL;
	FILE* rc_dump = NULL;
	int idle_size = 0;
	int configs = argc - PARAMS_MIN ;
	int drain_cfg = 0;
	int src_cfg = configs-1;
	int configured = 0;
	int config_size = 0;
	
	struct task_info task;
	task.current = NULL;
	
	if ( argc < PARAMS_MIN ) {
		fprintf(stderr,"To few arguments passed to simulator\n");
		fprintf(stderr,"%s x-size y-size stats-file io-dump idle-size config-size [kernels]\n",argv[0]);
		return -1;
	} else if ( !parse_arguments(argc,argv,&x,&y,&stats_file,&io_file,&idle_size,&config_size) ) {
		return -1;
	}
	
	if ( !initialize_memory(&memory_base) ) {
		return -1;
	}
	if ( !initialize_io_dump(&io_dump,io_file) ) {
		return -1;
	}
	if ( !initialize_rc_dump(&rc_dump) ) {
		return -1;
	}
	if ( !grid_create(x,y,&gstart) ) {
		return -1;
	}
	if ( !rc_managers_create(gstart,memory_base,x,&rc_list) ) {
		rc_managers_free(&rc_list);
		return -1;
	}
	if ( !memory_grid_connection(&mem_units,gstart,memory_base,idle_size) ) {
		return -1;
	}
	if ( !initialize_records(stats_file,&fd,x,y) ) {
		return -1;
	}
	
	reset_tick_record(&cnt);
	memset(memory_base,0xEE,MEMORY_SIZE);
	if ( !load_libraries(argc,argv,x,y,memory_base,&libs) ) {
		return -1;
	}
	
	for ( int i = 0 ; i < idle_size ; i++ ) {
		fprintf(rc_dump,"000 ");
		print_address(rc_dump,0);
		print_address(rc_dump,0);
		fprintf(rc_dump,"\n");
	}
	
#ifdef DEBUG
	printf("========== Initial state ========\n");
	print_grid(gstart);
	draw_complete_grid(gstart,mem_units);
	print_rc_managers(rc_list);
	print_memory_rc_managers(mem_units);
	print_memory_units(mem_units);
#endif
	for ( int i = 0 ; i < SWITCH_RATE(x,y)*(20*argc); i++,global_ticks++) {
		reset_tick_record(&cnt);
		if ( global_ticks < 0 ) {
		
		} else if ( global_ticks % SWITCH_RATE(x,y) == 0 ) {
			printf("======== TASK SWITCH ======\n");
			switch_tasks(x,y,&task,libs,rc_list,mem_units,memory_base);
		}
		if ( global_ticks % SWITCH_RATE(x,y) == 0 && global_ticks % DATA_TRANSFER_TICKS != 0 ) {
			fprintf(stderr,"Error: Offset task switch\n");
			break;
		}
		if ( global_ticks % DATA_TRANSFER_TICKS == 0 ) {
			
			if ( global_ticks % SWITCH_RATE(x,y) == 0 && !(configured == 1 && argc == (PARAMS_MIN+1))) {
				//fprintf(rc_dump,"DDD\n");
				if ( configured == 0 ) {
					print_rc(rc_dump,"001",src_cfg*y*config_size,drain_cfg*y*config_size,0);
					configured = 1;
				} else {
					print_rc(rc_dump,"011",src_cfg*y*config_size,drain_cfg*y*config_size,0);
				}
				
				drain_cfg = src_cfg;
				if ( src_cfg == 0 ) {
					src_cfg = configs-1;
				} else {
					src_cfg--;
				}
				
			} else {
				print_rc(rc_dump,"000",0,0,idle_size);
			}
			
		}
		printf("========= TICK %2d ======\n",global_ticks);
		if ( global_ticks % DATA_TRANSFER_TICKS == 0) {
			printf("========= DATA TRANSFER TICK =====\n");
			memory_grid_transfer(mem_units);
			grid_internal_register_transfer(gstart);
		}
		
		memory_rc_tick(mem_units,&cnt);
		memory_unit_tick(mem_units,&cnt,idle_size);
		rc_tick(rc_list,&cnt);
		grid_tick(gstart,&cnt);
#ifdef DEBUG
		printf("========== State %d ========\n",global_ticks);
		dump_memory(memory_base);
		print_grid(gstart);
#endif
		draw_complete_grid(gstart,mem_units);

		print_rc_managers(rc_list);
		print_memory_rc_managers(mem_units);
		print_memory_units(mem_units);
#ifdef DEBUG
		print_counters(&cnt);
#endif
		save_tick_record(fd,global_ticks,&cnt);
	}
#ifdef DEBUG
	print_grid(gstart);
	draw_grid(gstart);
	dump_memory(memory_base);
	draw_grid(gstart);
	draw_complete_grid(gstart,mem_units);
#endif
	perform_io_dump(io_dump,rc_dump,mem_units,idle_size);
	
	unload_configuration(&libs);
	grid_free(&gstart);
	rc_managers_free(&rc_list);
	free_memory_connection(&mem_units);
	free_memory(&memory_base);
	finalize_records(&fd);
	fclose(rc_dump);
	fclose(io_dump);
	return 0;
}

static int load_libraries(int argc, char** argv, int x, int y,
			uint8_t* memory_target,
			struct libraries** list) {
	int i = PARAMS_MIN; /* Skip executable name and other parameters */
	
	for ( ; i < argc; i++ ) {
		printf("Loading library: %s\n",argv[i]);
		if ( !load_configuration(argv[i],x,y,memory_target,list) ) {
			printf("Load configuration failed\n");
			return 0;
		}
	}
	return 1;
}


static void switch_tasks(int x, int y,
		struct task_info* task_info, 
		struct libraries* list,
		struct rc_manager* rc_list,
		struct mem_interface_rc* mem_units,
		uint8_t* memory_base) {
	uint32_t drain = 0;
	uint32_t source = 0;
	struct libraries* previous = task_info->current;
	
	if ( list == NULL ) {
		printf("Unable to switch task. None found\n");
		return;
	}
	
	/* No current task */
	if ( task_info->current == NULL ) {
		/* Generate dummy flush address*/
		drain = find_free_task_space(list,x,y,memory_base) 
			- memory_base;
		printf("Flushing grid to %08X\n",drain);
		task_info->current = list;
	} else {
		if ( list->next == NULL ) {
			/* Single task system */
			return;
		}
		/* Check if we have come to the end of the list */
		if ( task_info->current->next == NULL ) {
			task_info->current = list;
		} else {
			task_info->current = task_info->current->next;
		}
		drain = previous->load_address-memory_base;
	}
	
	if ( previous == task_info->current || task_info->current == NULL) {
		fprintf(stderr,"ERROR: No progress in task switch\n");
		return;
	}
	
	source = task_info->current->load_address - memory_base;
	
	printf("Switching Task\n");
	printf("New: %s\n",task_info->current->file_name);
	if ( previous != NULL) {
		printf("Old: %s\n",previous->file_name);
	}
	printf("source: %08x, drain: %08x\n",source,drain);
	
	memory_rc_setup(mem_units,x,y,source,drain);
	
	rc_setup(rc_list,source + memory_configuration_size(x,y)
		,drain+memory_configuration_size(x,y)
	);
}

static int parse_arguments(int argc, char** argv, int* x, int* y, 
			char** stats_file, char** io_file,
			int* idle_size, int* config_size) {
	char *end;
	errno = 0;
	
	*x = (int)strtol(argv[1],&end,10);
	
	printf("X size : %d\n",*x);
	if ( end == argv[1] || end != argv[1] + strlen(argv[1])) {
		fprintf(stderr,"unable to parse x\n");
		return 0;
	} else if ( errno == ERANGE || errno == EINVAL) {
		perror("x");
		return 0;
	}
	
	errno = 0;
	
	*y = (int)strtol(argv[2],&end,10);
	
	printf("Y size : %d\n",*y);
	if ( end == argv[2] || end != argv[2] + strlen(argv[2])) {
		fprintf(stderr,"unable to parse y\n");
		return 0;
	} else if ( errno == ERANGE || errno == EINVAL ) {
		perror("y");
		return 0;
	}

	*stats_file = argv[3];
	*io_file = argv[4];
	
	errno = 0;
	
	*idle_size = (int)strtol(argv[5],&end,10);
	
	printf("Idle size : %d\n",*idle_size);
	if ( end == argv[5] || end != argv[5] + strlen(argv[5])) {
		fprintf(stderr,"unable to parse idle size\n");
		return 0;
	} else if ( errno == ERANGE || errno == EINVAL ) {
		perror("idle size");
		return 0;
	}
	
	*config_size = (int)strtol(argv[6],&end,10);
	
	printf("Config size : %d\n",*y);
	if ( end == argv[6] || end != argv[6] + strlen(argv[6])) {
		fprintf(stderr,"unable to parse idle size\n");
		return 0;
	} else if ( errno == ERANGE || errno == EINVAL ) {
		perror("config size");
		return 0;
	}

	return 1;
}

static void print_rc(FILE* rc_dump, char* ctrl, 
			int source_address, int drain_address,
			int idles) {
	int i = 0;
	
	fprintf(rc_dump,"%s ",ctrl);
	print_address(rc_dump,source_address);
	print_address(rc_dump,drain_address);
	fprintf(rc_dump,"\n");
	for ( ; i < idles; i++ ) {
		fprintf(rc_dump,"000 ");
		print_address(rc_dump,0);
		print_address(rc_dump,0);
		fprintf(rc_dump,"\n");
	}
}

static void print_address(FILE* target, int address) {
	int i = ADDRESS_SIZE -1;
	
	for ( ; i >= 0 ; i-- ) {
		fprintf(target,"%c", ((address>>i)&0x01)+'0');
	}
	fprintf(target," ");
}

static void perform_io_dump(FILE* target, FILE* rc_data,
			struct mem_interface_rc* mem_units,
			int idle_size) {
	fseek_rc(rc_data);
	private_data_fseek(mem_units);
	while ( !stripe_exists(mem_units) ) {
		drain_rc(target,rc_data);
		drain_stripe(mem_units,idle_size,target);
		fprintf(target,"\n");
	}

	
}

static void fseek_rc(FILE* rc_data) {
	int i = 0;
	int current = '\0';	
	fseek(rc_data,0,SEEK_SET);
	while ( (current = fgetc(rc_data)) != EOF ) {
		if ( current == '\n' ) {
			i++;
		}
		if ( i == PRE_TRIGGER ) {
			break;
		}
	}
}

static void drain_rc(FILE* target, FILE* rc_data) {
	int current = '\0';
	int read = 0;
	
	while ( (current = fgetc(rc_data)) != EOF ) {
		if ( current == '\n' ) {
			
			break;
		}
		read = 1;
		fputc(current,target);
	}
	/* Output dummy data. End of output is upon us */
	if ( read == 0 ) {
		fprintf(target,"000 ");
		print_address(target,0);
		print_address(target,0);
	}
}

static int initialize_io_dump(FILE** io_dump, char* file_name) {
	if ( io_dump == NULL ) {
		return 0;
	}
	*io_dump = fopen(file_name,"w");
	if ( *io_dump == NULL ) {
		perror("initialize io dump:");
		return 0;
	}
	return 1;
}

static int initialize_rc_dump(FILE** rc_dump) {
	if ( rc_dump == NULL ) {
		return 0;
	}
	*rc_dump = tmpfile();
	if ( *rc_dump == NULL ) {
		perror("initialize rc dump:");
		return 0;
	}
	return 1;
}

