#ifndef _MEMORY_H_
#define _MEMORY_H_

#include "grid.h"
#include "counters.h"


#define MEMORY_SIZE (2<<24) /* 2048b */
#define MEMORY_RC_TICKS DATA_TRANSFER_TICKS
#define MEMORY_ACCESS_TICKS DATA_TRANSFER_TICKS
#define MEMORY_ACCESS_SIZE (DATA_PORT_SIZE/MEMORY_ACCESS_TICKS)

#ifndef MEMORY_RC_SIZE
	#warning "Default memory rc size used"
	#define MEMORY_RC_SIZE (128) /* 128 bit configuration */
#endif
#define MEMORY_RC_ACCESS_SIZE (MEMORY_RC_SIZE/MEMORY_RC_TICKS) // (sizeof(struct mem_pattern)/MEMORY_RC_TICKS)

enum mem_config_state {
	MEM_UNCONFIGURED,
	MEM_RECONFIGURING,
	MEM_FETCHING,
	MEM_STALLED
};

enum mem_interface_type {
	FROM_MEM,
	TO_MEM,
	UNCONFIGURABLE_MEM
};

static const char* mem_state_string[4] = {
	"Unconfigured",
	"Reconfiguration",
	"Fetching",
	"Stalled"
};

struct mem_pattern {
	uint32_t address; /* Offset in main memory */
	uint32_t address_limit;
	uint32_t stride;
	uint32_t dummy; /* Added for nice alignment in memory printout */
};

struct mem_rc_bus {
	struct mem_pattern source;
	struct mem_pattern drain;
};

struct mem_interface{
	enum mem_config_state state;
	enum mem_interface_type type;
	int ticks_left;
	struct grid_cell* connection;
	cell_valid output_valid;
	cell_transfer output;
	cell_valid* ioport_valid;
	cell_transfer* ioport;
	uint8_t* memory_base;
	struct mem_pattern mem_info;
	struct mem_rc_bus* rc_bus;
	struct mem_interface* next;
	FILE* private_dump;
	int current_idle;
};

/* TODO: Unify with rc_managers */
enum mem_rc_state {
	MEM_RC_IDLE,
	MEM_RC_PRESTART,
	MEM_RC_DELAYED,
	MEM_RC_ACTIVE
};

static const char* mem_rc_state_string[4] = {
	"Idle",
	"Prestart",
	"Delayed",
	"Active"
};

struct mem_interface_rc {
	enum mem_rc_state state;
	int ticks_left;
	uint32_t source_address;
	uint32_t drain_address;
	uint8_t* memory_base;
	struct mem_rc_bus rc_bus;
	struct mem_interface* last;
	struct mem_interface* munit_list;
	struct mem_interface_rc* next;
};

int initialize_memory(uint8_t** memory_base);
void free_memory(uint8_t** memory_base);
int memory_grid_connection(struct mem_interface_rc** list, 
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle);
void free_memory_connection(struct mem_interface_rc** list);
void memory_unit_tick(struct mem_interface_rc* list,
			struct counters* cnt,
			int idle_size);
void memory_rc_tick(struct mem_interface_rc* list,
			struct counters* cnt);
void memory_rc_setup(struct mem_interface_rc* list,
			int x_size, int y_size,
			uint32_t source_memory_address,
			uint32_t drain_memory_address);
void memory_grid_transfer(struct mem_interface_rc* list);
void private_data_fseek(struct mem_interface_rc* list);
void drain_stripe(struct mem_interface_rc* list, 
				int idles, FILE* target);
int stripe_exists(struct mem_interface_rc* list);

#endif /* !_MEMORY_H_ */
