#include <stdio.h>
#include <ctype.h>

#include "print_utils.h"
#include "grid.h"
#include "cell.h"
#include "rc_manager.h"
#include "memory.h"
#include "counters.h"

static void print_cell_top(struct grid_cell* gcell);
static void print_cell_header(struct grid_cell* gcell);
static void print_cell_footer(struct grid_cell* gcell);
static void print_cell_bottom(struct grid_cell* gcell);
static void print_memory_top(void);
static void print_memory_header(struct mem_interface* list);
static void print_memory_footer(struct mem_interface* list);
static void print_memory_bottom(void);

static void print_memory_list(struct mem_interface* list);


void print_grid(struct grid_cell* NW_corner) {
	struct grid_cell* current_row = NW_corner;
	struct grid_cell* current_col = NULL;

	/* Traverse the grid */
	while ( current_row != NULL ) {
		current_col = current_row;
		while ( current_col != NULL ) {
			print_grid_cell(current_col);
			current_col = current_col->east;
		}
		current_row = current_row->south;
	}
}

void print_grid_cell(struct grid_cell* gcell) {
	printf("----------------------\n");
	printf("X: %d\t Y: %d\n",gcell->x,gcell->y);
	print_cell(&(gcell->cell));
	return;
}

void print_cell(cell_info* cell) {
	printf("Cell %p\n", cell);
	printf("Config_state: %s\n", cell_config_state_string[cell->config_state]);
	printf("Ticks left: %d\n",cell->ticks_left);
	printf("Configuration: %p\n", cell->comp_state.fabric);
	printf("State_register: %lld 0x%llX\n",cell->comp_state.state,cell->comp_state.state);
	printf("Configuration bus: %p\n",cell->config_bus);
	printf("Inputs:\n");
	printf("North:\t%lld\n",cell->register_north);
	printf("West:\t%lld\n", cell->register_west);
	printf("Outputs\n");
	printf("East:\t%lld\n", cell->result_east);
	printf("South:\t%lld\n", cell->result_south);
	return;
}

void draw_grid(struct grid_cell* NW_corner) {
	struct grid_cell* current_row = NW_corner;
	struct grid_cell* current_col = NULL;

	/* Traverse the grid */
	while ( current_row != NULL ) {
		current_col = current_row;
		printf("-");
		while ( current_col != NULL ) {
			print_cell_top(current_col);
			current_col = current_col->east;
		}
		printf("\n|");
		current_col = current_row;
		while ( current_col != NULL ) {
			print_cell_header(current_col);
			current_col = current_col->east;
		}
		printf("\n|");
		current_col = current_row;
		while ( current_col != NULL ) {
			print_cell_footer(current_col);
			current_col = current_col->east;
		}
		if ( current_row->south == NULL ) {
			printf("\n-");
			current_col = current_row;
			while ( current_col != NULL ) {
				print_cell_bottom(current_col);
				current_col = current_col->east;
			}
		} else {
			printf("\n");
		}
		current_row = current_row->south;
	}
	printf("\n");
}

static void print_cell_top(struct grid_cell* gcell) {
	printf("------");
}

static void print_cell_header(struct grid_cell* gcell) {
	printf("%c    |",*cell_config_state_string[gcell->cell.config_state]);
}

static void print_cell_footer(struct grid_cell* gcell) {
	printf("     |");
}

static void print_cell_bottom(struct grid_cell* gcell) {
	printf("------");
}

void print_rc_managers(struct rc_manager* list) {
	struct rc_manager* cur = list;
	
	while ( cur != NULL ) {
		printf("-----------\n");
		printf("RC: %p\n",cur);
		printf("Stride: %d\n",cur->stride);
		printf("Source memory address: 0x%X\n", 
			cur->source_memory_address
		);
		printf("Drain memory address: 0x%X\n",
			cur->drain_memory_address
		);
		printf("Memory base: %p\n",cur->memory_base);
		printf("State: %s\n",rc_state_string[cur->state]);
		printf("Ticks: %d\n",cur->ticks_left);
		
		cur = cur->next;
	}
}

void dump_memory(uint8_t* memory_base) {
	int i = 0;
	int j = 0;
	printf("%p:\n",memory_base);
	for ( ;i < MEMORY_SIZE; i += 16) {
		printf("%6X",i);
		for ( j = 0 ; j < 16; j++ ) {
			if ( j+i < MEMORY_SIZE ) {
				printf(" %02X",*(memory_base+i+j));
			}
		}
		printf("\t");
		for ( j = 0 ; j < 16; j++ ) {
			if ( j+i < MEMORY_SIZE ) {
				if (isprint(*(memory_base+i+j))) {
					printf("%c",*(memory_base+j+i));
				} else {
					printf(".");
				}
			}
		}
		printf("\n");
	}

}

void print_memory_rc_managers(struct mem_interface_rc* list) {
	struct mem_interface_rc* cur = list;
	
	while ( cur != NULL ) {
		printf("----------\n");
		printf("Mem_rc: %p\n",cur);
		printf("State: %s\n",mem_rc_state_string[cur->state]);
		printf("Source memory address: 0x%X\n", cur->source_address);
		printf("Drain memory address: 0x%X\n", cur->drain_address);
		printf("Memory base: %p\n",cur->memory_base);
		printf("Last: %p\n",cur->last);
		printf("Ticks: %d\n",cur->ticks_left);
		cur = cur->next;
	}
}

void print_memory_units(struct mem_interface_rc* list) {
	struct mem_interface_rc* cur = list;
	
	while ( cur != NULL ) {
		print_memory_list(cur->munit_list);
		cur = cur->next;
	}
	
}

static void print_memory_list(struct mem_interface* list) {
	struct mem_interface* cur = list;
	
	while ( cur != NULL ) {
		printf("----------\n");
		printf("Memory unit: %p\n", cur);
		printf("State: %d\n",cur->state);
		printf("Type: %d\n",cur->type);
		printf("Ticks left: %d\n",cur->ticks_left);
		printf("Connected to cell %p\n",&(cur->connection->cell));
		printf("Configuration bus: %p\n",cur->rc_bus);
		printf("Output  : 0x%llX\n",cur->output);
		printf("IO-port : %p\n",cur->ioport);
		printf("MEM-base: %p\n",cur->memory_base);
		printf("Address:  %lX\n",cur->mem_info.address);
		printf("Alimit:   %lX\n",cur->mem_info.address_limit);
		printf("Stride:   %lX\n",cur->mem_info.stride);
		
		cur = cur->next;
	}
}

void draw_complete_grid(struct grid_cell* NW_corner, 
			struct mem_interface_rc *mem) {
	struct mem_interface* munit_list_top = NULL;
	struct mem_interface* munit_list_west = NULL;
	struct mem_interface* munit_list_top_start = NULL;
	struct mem_interface* munit_list_top_tail = NULL;
	struct mem_interface* munit_list_west_start = NULL;
	struct mem_interface* munit_list_west_tail = NULL;
	struct grid_cell* current_row = NW_corner;
	struct grid_cell* current_col = NULL;
	
	
	/* Try to detect toplist.. should always be the first one in the list */
	if ( mem->munit_list != NULL && mem->munit_list->type == FROM_MEM ) {
		munit_list_top = mem->munit_list;
		if ( mem->next != NULL ) {
			munit_list_west = mem->next->munit_list;
		}
	} else if ( mem->next != NULL && mem->next->munit_list != NULL ) {
		munit_list_top = mem->next->munit_list;
		munit_list_west = mem->munit_list;
	}
	munit_list_top_start = munit_list_top;
	
	/* Draw memoryheader */
	printf("      M"); /* Spacer */
	munit_list_top = munit_list_top_start;
	while(munit_list_top != NULL && munit_list_top->type == FROM_MEM) {
		
		print_memory_top();
		munit_list_top = munit_list_top->next;
	}
	printf("\n      |");
	munit_list_top = munit_list_top_start;
	while(munit_list_top != NULL && munit_list_top->type == FROM_MEM) {
		
		print_memory_header(munit_list_top);
		munit_list_top = munit_list_top->next;
	}
	printf("\n      |");
	munit_list_top = munit_list_top_start;
	while(munit_list_top != NULL && munit_list_top->type == FROM_MEM) {
		
		print_memory_footer(munit_list_top);
		munit_list_top = munit_list_top->next;
	}
	
	/* Spool toplist */
	munit_list_top_tail = munit_list_top_start;
	while ( munit_list_top_tail != NULL 
			&& munit_list_top_tail->type != UNCONFIGURABLE_MEM) {
		munit_list_top_tail = munit_list_top_tail->next;
	}
	/* Skip separator */
	if ( munit_list_top_tail != NULL ) {
		munit_list_top_tail = munit_list_top_tail->next;
	}
	printf("\n");
	
	munit_list_west_start = munit_list_west;
	munit_list_top = munit_list_top_tail;
	
	while ( current_row != NULL ) {
		current_col = current_row;
		
		print_memory_top();
		printf("-");
		while ( current_col != NULL ) {
			print_cell_top(current_col);
			current_col = current_col->east;
		}
		print_memory_top();
		printf("\n|");
		print_memory_header(munit_list_west);
		current_col = current_row;
		while ( current_col != NULL ) {
			print_cell_header(current_col);
			current_col = current_col->east;
		}
		print_memory_header(munit_list_top);
		printf("\n|");
		print_memory_footer(munit_list_west);
		current_col = current_row;
		while ( current_col != NULL ) {
			print_cell_footer(current_col);
			current_col = current_col->east;
		}
		print_memory_footer(munit_list_top);
		
		if ( current_row->south == NULL ) {
			/* Last grid_row */
		} else {
			printf("\n");
		}
		current_row = current_row->south;
		munit_list_west = munit_list_west->next;
		munit_list_top = munit_list_top->next;
	}
	printf("\n-");
	
	/* Spool westlist */
	munit_list_west_tail = munit_list_west_start;
	while ( munit_list_west_tail != NULL 
			&& munit_list_west_tail->type != UNCONFIGURABLE_MEM) {
		munit_list_west_tail = munit_list_west_tail->next;
	}
	if ( munit_list_west_tail != NULL ) {
		munit_list_west_tail = munit_list_west_tail->next;
	}
	/* Draw memoryheader */
	printf("------"); /* Spacer */
	munit_list_west = munit_list_west_tail;
	while(munit_list_west != NULL && munit_list_west->type == TO_MEM) {
		print_memory_bottom();
		munit_list_west = munit_list_west->next;
	}
	printf("------"); /* Spacer */
	printf("\n      |");
	munit_list_west = munit_list_west_tail;
	while(munit_list_west != NULL && munit_list_west->type == TO_MEM) {
		
		print_memory_header(munit_list_west);
		munit_list_west = munit_list_west->next;
	}

	printf("\n      |");
	munit_list_west = munit_list_west_tail;
	while(munit_list_west != NULL && munit_list_west->type == TO_MEM) {
		
		print_memory_footer(munit_list_west);
		munit_list_west = munit_list_west->next;
	}

	printf("\n      M");
	munit_list_west = munit_list_west_tail;
	while(munit_list_west != NULL && munit_list_west->type == TO_MEM) {
		
		print_memory_top();
		munit_list_west = munit_list_west->next;
	}
	printf("\n");

}

static void print_memory_top() {
	printf("MMMMMM");
}

static void print_memory_header(struct mem_interface* mem){
	printf("%c    |",*mem_state_string[mem->state]);
}
static void print_memory_footer(struct mem_interface* mem){
	printf("     |");
}

static void print_memory_bottom(){
	printf("------");
}

void print_counters(struct counters* cnt) {
	printf("Total memory accesses:    %lld\n",cnt->total_memory_accesses);
	printf("Tick memory accesses:     %lld\n",cnt->tick_memory_accesses);
	printf("Total rc memory accesses: %lld\n",cnt->total_rc_memory_accesses);
	printf("Tick rc memory accesses:  %lld\n",cnt->tick_rc_memory_accesses);
	printf("Total rc ticks:           %lld\n",cnt->total_rc_ticks);
	printf("Tick rc ticks:            %lld\n",cnt->tick_rc_ticks);
	printf("Total computation ticks:  %lld\n",cnt->total_computation_ticks);
	printf("Tick computation ticks:   %lld\n",cnt->tick_computation_ticks);
}
