#ifndef _RC_MANAGER_H_
#define _RC_MANAGER_H_
#include <stdint.h>
#include "cell.h"
#include "grid.h"
#include "counters.h"

#define RC_TICKS DATA_TRANSFER_TICKS
#ifndef RC_SIZE
	#warning "Default rc size used"
	#define RC_SIZE (128) /* 128 bit configuration */
#endif
#define RC_BUS_WIDTH (RC_SIZE/RC_TICKS) //(sizeof(struct cell_rc_targets)/RC_TICKS)

enum rc_state {
	IDLE,
	PRESTART,
	ACTIVE
};

static const char* rc_state_string[3] = {
	"Idle", 
	"Prestart", 
	"Active"
};

/* This should be possible by just using two circular registers 
 * in hw. But.. that would probably slow the system down */
/* TODO: Make this two cell_rc_target structs */
struct rc_bus {
	struct cell_rc_targets source;
	struct cell_rc_targets drain;
};

struct rc_manager {
	int ticks_left;
	int stride; /* Is source and drain stride always equal ? */
	uint32_t source_memory_address;
	uint32_t drain_memory_address;
	uint8_t* memory_base;
	enum rc_state state;
	struct rc_bus bus;
	struct grid_cell* rc_list;
	struct grid_cell* last;
	struct rc_manager* next;
};

int rc_managers_create(struct grid_cell* NW_corner, 
			uint8_t* memory_base, 
			int x_size, 
			struct rc_manager** list);
void rc_managers_free(struct rc_manager** list);
void rc_setup(struct rc_manager* list, 
			uint32_t source_memory_address, 
			uint32_t drain_memory_address);
void rc_tick(struct rc_manager* list,
			struct counters* cnt);


#endif /* !_RC_MANAGER_H_ */