library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_textio.all;
use ieee.math_real.all;
use std.textio.all;
use work.common_defs.all;

entity rc_master_rom is
	generic(
		elements	:positive	:= 4;
		data_width	:positive	:= 6;
		init_source	:string		:= "roms/default"
	);
	port (
		addr	: in std_logic_vector(integer(ceil(log2(real(elements))))-1 downto 0);
		output	: out std_logic_vector(data_width-1 downto 0)
	);
end entity;

architecture rcm_rom of rc_master_rom is
	type rom_type	is array(2**integer(ceil(log2(real(elements))))-1 downto 0) of std_logic_vector(data_width-1 downto 0);
	
	-- Modified version. Orignal from the XST manual (ug627)
	impure function InitRcmRomFromFile(RomFileName : in string) return rom_type is
		File RomFile		: text is in RomFileName;
		variable RomFileLine	: line;
		variable ROM		: rom_type;
	begin
		for I in rom_type'low to rom_type'high loop
			readline (RomFile, RomFileLine);
			read (RomFileLine, ROM(I));
		end loop;
		return ROM;
	end function;
	
	signal data_rom : rom_type := InitRcmRomFromFile(init_source
		& "_"
		& integer'image(elements)
		& "_"
		& integer'image(data_width)
		& ".txt");
	
begin
	process(addr,data_rom)
	begin
		output <= data_rom(conv_integer(addr));
	end process;
end architecture;
