library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;

entity runtime_manager is
	generic(
		axis		:natural	:= 4
	);
	port(
		clk		:in std_logic;
		reset		:in std_logic;
		
		override	:in std_logic;
		rc_active	:in std_logic;
		vertical	:in std_logic;
		idle_out	:out std_logic
	);

end entity;

architecture rt_manager of runtime_manager is 
	signal idle_pattern	: std_logic_vector(axis-1-2 downto 0);
	signal output_buffer	: std_logic_vector(1 downto 0);
begin
	
	idle: process(clk,reset) is
	
	begin
		if reset = '0' then
			idle_pattern <= (others => '0');
			output_buffer <= (others => '0');
		elsif rising_edge(clk) then
			-- Shift data
			idle_pattern(idle_pattern'high-1 downto 0) <= idle_pattern(idle_pattern'high downto 1); 
			output_buffer(0) <= output_buffer(1);
			idle_pattern(idle_pattern'high) <= output_buffer(0);
			
			-- Change pattern 
			if rc_active = '1' then
				output_buffer(output_buffer'high) <= not (vertical or override);
			else 
				output_buffer(output_buffer'high) <= idle_pattern(0);
			end if;
		end if;
	end process;
	idle_out <= output_buffer(0);
end architecture;