library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;

entity shift_register is 
	generic(
		width		: positive	:= 8
	);
	port(
		clk		: in std_logic;
		reset		: in std_logic;
		
		write		: in std_logic;
		
		shift_left 	: in std_logic;
		shift_right	: in std_logic;
		
		data_in		: in std_logic_vector(width-1 downto 0);
		data_out	: out std_logic_vector(width-1 downto 0);
		
		carry_in	: in std_logic;
		carry_out	: out std_logic
	);
end entity;

architecture shifter of shift_register is 
	signal	shift_reg : std_logic_vector(width-1 downto 0);
begin
	
	process(clk,reset) 
	begin
		if reset = '0' then
			-- Clear register
			shift_reg <= (shift_reg'range => '0');
			carry_out <= '0';
		elsif rising_edge(clk) and write = '1' then
			if shift_left = '1' then
				shift_reg(width-1 downto 1) <= data_in(data_in'left -1 downto 0);
				shift_reg(0) <= carry_in;
				carry_out <= data_in(data_in'left);
			elsif shift_right = '1' then
				shift_reg(width-1) <= carry_in;
				shift_reg(width-2 downto 0) <= data_in(data_in'left downto 1);
				carry_out <= data_in(0);
			else
				shift_reg <= data_in;
				carry_out <= '0';
			end if;
			--shift_reg <= data_in;
		end if;
	end process;
	
	data_out <= shift_reg;
end architecture;