library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.log2;
use work.common_defs.all;

entity shifter is 
	generic(
		width		:positive := 32
	);
	port (
		data		:in	std_logic_vector(width -1 downto 0);
		
		shift_amount	:in	std_logic_vector(width -1 downto 0);
		direction	:in	std_logic;
		
		output		:out	std_logic_vector(width -1 downto 0)
	);
end entity;

architecture logic_shifter_mux of shifter is
	signal shift_bits	: std_logic_vector(integer(log2(real(width)))-1 downto 0);
	signal force_zero	: std_logic;
begin
	shift_bits <= shift_amount(shift_bits'left downto 0);
	-- Catch large shiftoperands
	force_zero <= '0' when (shift_amount(shift_amount'left downto shift_bits'left+1) = (shift_amount'left downto shift_bits'left +1=> '0')) else '1';
	
	shift : process(data, shift_bits, direction, force_zero) 
	variable pre_stage	: std_logic_vector(width -1 downto 0);
	variable post_stage	: std_logic_vector(width -1 downto 0);
	begin
		pre_stage := data;
		post_stage := pre_stage;
		for I in shift_bits'range loop
			if shift_bits(I) = '1' then
				if direction = '1' then
					post_stage := pre_stage(width-1-(2**I) downto 0) & ((2**I)-1 downto 0 => '0');
				else 
					post_stage := ((2**I)-1 downto 0 => '0') & pre_stage(width-1 downto (2**I));
				end if;
			else 
				post_stage := pre_stage;
			end if;
			pre_stage := post_stage;
		end loop;
		if force_zero = '0' then
			output <= post_stage;
		else 
			output <= (output'range => '0');
		end if;
	end process;
end architecture;

