library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_addersubtracter_testbench is
end entity;

architecture addersubtracter_test of auto_addersubtracter_testbench is
	constant size		: integer := 32;

	signal operation_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal output_o		: std_logic_vector(size -1 downto 0);
	signal running		: std_logic := '1';
begin
	uut : addersubtracter
		generic map(
			width => size
		)
		port map(
			data_a => data_a_i,
			data_b => data_b_i,
			operation => operation_i,
			output => output_o
		);
	
	random_test : process is
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-addersubtracter.txt";
			variable L: line;
			
			variable operation	:std_logic;
			variable first_operand	:integer;
			variable second_operand	:integer;
			variable result		:integer;
		
			--variable op		:integer;
		
			procedure perform_simulation is 
		
			begin
				operation_i  <= operation;
				data_a_i  <= std_logic_vector(to_signed(first_operand,size));
				
				data_b_i  <= std_logic_vector(to_signed(second_operand,size));
				
				wait for 100 ns; -- shouldn't this be valid at the same time ?
				assert output_o = std_logic_vector(to_signed(result,size)) report "Incorrect result" severity failure;
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,operation);
				read(L,first_operand);
				read(L,second_operand);
				read(L,result);
			
				perform_simulation;
			end loop;
		end procedure;
	begin
		data_a_i <= (data_a_i'range => '0');
		data_b_i <= (data_b_i'range => '0');
		operation_i <= '0';
		random_stimuli;
		-- End of testbench
		wait;
	end process;

end architecture;