#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#define op_t int32_t
#define OP_LENGTH (sizeof(op_t)*8)
#define OPCODE_LENGTH 3
#define FLAGS_LENGTH 2
static void write_opcode(FILE* fd, char operation);
static void write_operand(FILE* fd, op_t op);
static void write_flags(FILE* fd, op_t result);


int main(int argc, char** argv) {
	long int vectors = 0;
	FILE* fd = NULL;
	char operation;
	op_t operand_a,operand_b,result;
	
	time_t timeseed = time(NULL);
	srand(timeseed);
	
	if ( argc != 3 ) {
		printf("Usage: %s vectors <output filname>\n",argv[0]);
		return -1;
	}
	
	errno = 0;
	vectors = strtol(argv[1],NULL,10);
	if ( errno == ERANGE || errno == EINVAL ) {
		fprintf(stderr,"Parse error: %s\n",strerror(errno));
		return -1;
	}
	
	if ( (fd = fopen(argv[2],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	for (;vectors > 0 ; vectors-- ) {
		operation = (rand()%8);
		operand_a = rand();
		operand_b = rand();
		
		if ( rand()%4 == 0 ) {
			operand_a = operand_a * -1;
		}
		if ( rand()%4 == 0 ) {
			operand_b = operand_b * -1;
		}
		
		switch (operation) {
		case 0:
			/* ADD */
			result = operand_a + operand_b;
			break;
		case 1:
			/* SUB */
			result = operand_a - operand_b;
			break;
		case 2:
			/* SRL */
			if ( ((uint32_t)operand_b) >= 32) {
				operand_a = 0;
			}
			result = (uint32_t)operand_a >> (uint32_t)operand_b;
			break;
		case 3:
			/* SLL */
			if ( ((uint32_t)operand_b) >= 32) {
				operand_a = 0;
			}
			result = (uint32_t)operand_a << (uint32_t)operand_b;
			break;
		case 4:
			/* OR */
			result = operand_a | operand_b;
			break;
		case 5:
			/* NOR */
			result = ~(operand_a | operand_b);
			break;
		case 6:
			/* AND */
			result = operand_a & operand_b;
			break;
		case 7:
			/* XOR */
			result = operand_a ^ operand_b;
			break;
		default:
			fprintf(stderr,"Invalid operation\n");
			return -1;
		}
		
		write_opcode(fd,operation);
		write_operand(fd,operand_a);
		write_operand(fd,operand_b);
		write_operand(fd,result);
		write_flags(fd,result);
		fprintf(fd,"\n");
	}
	fclose(fd);

	return 0;
}

static void write_opcode(FILE* fd, char operation) {
	int i = OPCODE_LENGTH -1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((operation>>i)&0x01));
	}
	fprintf(fd," ");
}

static void write_flags(FILE* fd, op_t result) {
	char flags = 0;
	int i = FLAGS_LENGTH -1;
	
	if ( result < 0 ) {
		flags = 0x02;
	} else if ( result == 0 ) {
		flags = 0x01;
	}
	
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((flags>>i)&0x01));
	}
	fprintf(fd," ");
}



static void write_operand(FILE* fd, op_t op) {
	int i = OP_LENGTH-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((op>>i)&0x01));
	}
	fprintf(fd," ");
}

