library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_multiplydivide_testbench is
end entity;

architecture auto_md_test of auto_multiplydivide_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal opcode_i		: std_logic;
	signal valid_a_i	: std_logic;
	signal valid_b_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal valid_output_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);
	signal running		: std_logic := '1';
begin
	uut : multiplydivider
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			opcode => opcode_i,
			valid_a => valid_a_i,
			valid_b => valid_b_i,
			data_a => data_a_i,
			data_b => data_b_i,
			valid_output => valid_output_o,
			output => output_o
		);
	
	
	clock_gen : process is
	begin
		while running='1' loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			valid_a_i <= '0';
			valid_b_i <= '0';
			reset <= '1';
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-multiplydivide.txt";
			variable L: line;
			
			variable first_operand	:integer;
			variable second_operand	:integer;
			variable result		:integer;
		
			variable div		:std_logic;
		
			procedure perform_simulation is 
		
			begin
				opcode_i  <= div;
				valid_a_i <= '1';
				data_a_i  <= std_logic_vector(to_signed(first_operand,size));
				
				valid_b_i   <= '1';
				data_b_i  <= std_logic_vector(to_signed(second_operand,size));
				
				wait until rising_edge(clk); -- Hack ?
				wait until valid_output_o = '1';
				wait for 1 ns; -- shouldn't this be valid at the same time ?
				assert output_o = std_logic_vector(to_signed(result,size)) report "Incorrect result" severity failure;
			
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,div);
				read(L,first_operand);
				read(L,second_operand);
				read(L,result);
				reset_uut;
				perform_simulation;
			end loop;
		end procedure;
	begin
		data_a_i <= (data_a_i'range => '0');
		data_b_i <= (data_b_i'range => '0');
		valid_a_i <= '0';
		valid_b_i <= '0';
		opcode_i <= '0';
		reset_uut;
		random_stimuli;
		-- Hack
		--assert false report "End of simulation" severity failure;
		running <= '0';
		wait for 500 ns;
		wait;
	end process;

end architecture;