library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_mux_testbench is
end entity;

architecture mux_test of auto_mux_testbench is
	constant size		: integer := 32;
	constant sels		: integer := 2; -- log2 of port numbers

	signal sel_i		: std_logic_vector(sels -1 downto 0);
	
	signal inputs_i		: std_logic_vector(((2**sels)*size) -1 downto 0);

	signal output_o		: std_logic_vector(size -1 downto 0);
begin
	uut : mux
		generic map(
			path_width => size,
			sel_width => sels
		)
		port map(
			sel => sel_i,
			input => inputs_i,
			output => output_o
		);
	
	random_test : process is
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-mux.txt";
			variable L: line;
			
			variable operation	:integer;
			variable first_operand	:integer;
			variable second_operand	:integer;
			variable third_operand	:integer;
			variable forth_operand	:integer;
			variable result		:integer;
		
		
			procedure perform_simulation is 
		
			begin
				sel_i  <= std_logic_vector(to_unsigned(operation,sels));
				inputs_i  <= std_logic_vector(to_signed(forth_operand,size))
					& std_logic_vector(to_signed(third_operand,size))
					& std_logic_vector(to_signed(second_operand,size))
					& std_logic_vector(to_signed(first_operand,size));
				
				wait for 100 ns;
				assert output_o = std_logic_vector(to_signed(result,size)) report "Incorrect result" severity failure;
			
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,operation);
				
				read(L,first_operand);
				read(L,second_operand);
				read(L,third_operand);
				read(L,forth_operand);
				read(L,result);
			
				perform_simulation;
			end loop;
		end procedure;
	begin
		sel_i <= (sel_i'range => '0');
		inputs_i <= (inputs_i'range => '0');
		random_stimuli;
		-- Hack
		wait;
	end process;

end architecture;