#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#define address_t uint16_t
/* 20 bits */
#define HEADER_SIZE 24
#define header_t uint32_t
#define control_t uint8_t
#define CELLS_X 8 
#define CELLS_Y 8
#define ADDRESS_DELTA 3
#define OFFSET_UNIT 1

#define FLAG_SIZE 4

static void write_control_t(FILE* fd, control_t value);
static void write_address_t(FILE* fd, address_t value);
static void write_header_t(FILE* fd, header_t value);
static void write_flags(FILE* fd, char value);
static int bits_set(control_t value);
static void generate_header(header_t* target,control_t* pattern, control_t* pattern_v, char* out_flags);
static control_t get_pattern(header_t size);

static void delay_write(char wrc, address_t waddr, address_t offset, control_t wpat, control_t wpatv,
		char* out_wrc, address_t* out_waddr, address_t* out_offset, control_t* out_wpat, control_t* out_wpatv);

int main(int argc, char** argv) {
	long int vectors = 0;
	FILE* fd = NULL;
	int steps = 0;
	char reconfiguring = 0;
	char configured = 0;
	char source_cf = 0;
	char drain_cf= 0;
	char reader_rc = 0;
	char writer_rc = 0;
	char rcm_write_en = 0;
	char write_allow = 0;
	char error = 0;
	char clear_error = 0;
	char flags = 0;
	header_t read_header = 0, write_header = 0;
	address_t src_base_address = 0, dest_base_address = 0;
	address_t rcm_read_address = 0, rcm_write_address = 0;
	address_t reader_addr = 0, reader_offset = 0;
	address_t writer_addr = 0, writer_offset = 0;
	control_t reader_pattern = 0, writer_pattern = 0;
	control_t reader_v_pattern = 0, writer_v_pattern = 0;
	
	char out_wrc = 0;
	address_t out_waddr = 0, out_offset = 0;
	control_t out_wpat = 0, out_wpatv = 0;
	
	
	time_t timeseed = time(NULL);
	srand(timeseed);
	
	if ( argc != 3 ) {
		printf("Usage: %s vectors <output filname>\n",argv[0]);
		return -1;
	}
	
	errno = 0;
	vectors = strtol(argv[1],NULL,10);
	if ( errno == ERANGE || errno == EINVAL ) {
		fprintf(stderr,"Parse error: %s\n",strerror(errno));
		return -1;
	}
	
	if ( (fd = fopen(argv[2],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	for (;vectors > 0 ; vectors-- ) {
		
		if ( reconfiguring  == 0 ) {
			reconfiguring = !(rand()%4);
			if ( reconfiguring == 1 ) {
				src_base_address = rand();
				dest_base_address = rand();
				rcm_read_address = src_base_address;
				rcm_write_address = dest_base_address;
				reader_addr = rcm_read_address + ADDRESS_DELTA;
				writer_addr = rcm_write_address + ADDRESS_DELTA;
				generate_header(&read_header,&reader_pattern,&reader_v_pattern,&flags);
				drain_cf = configured;
				source_cf = 1;
				write_allow = configured;
				writer_offset = bits_set(writer_v_pattern)*OFFSET_UNIT;
				reader_offset = bits_set(reader_v_pattern)*OFFSET_UNIT;
				if (reader_pattern == 0 || reader_v_pattern == 0 ) {
					error = 1;
					steps = 6;
				} else if (drain_cf == 1 && source_cf  == 0 ) {
					/* Skip two states when just draining */
					steps = CELLS_Y+4;
				} else {
					steps = CELLS_Y+6;
				}
			}
		} else {
			steps--;
			//source_cf = 0;
			//drain_cf = 0;
			if ( steps == CELLS_Y+1) {
				writer_rc = drain_cf;
				reader_rc = source_cf;
			} else if ( steps == 1 && error == 1 ) {
				clear_error = 1;
				reader_addr = 0;
				reader_offset = 0;
				reader_pattern = 0;
				reader_v_pattern = 0;
				flags = 0;
				drain_cf = 0;
				source_cf = 0;
			} else if ( steps == 1 && error == 0 ) {
				reconfiguring = 0;
				write_allow = 0;
				
				reader_rc = 0;
				writer_rc = 0;
				flags = 0;
				
				configured = source_cf;
				
				write_header = read_header;
				writer_addr = 0;
				writer_offset = reader_offset;
				writer_pattern = reader_pattern;
				writer_v_pattern = reader_v_pattern;
				
				reader_addr = 0;
				reader_offset = 0;
				reader_pattern = 0;
				reader_v_pattern = 0;
				drain_cf = 0;
				source_cf = 0;
			} else if ( steps ==  0 && error == 1) {
				reconfiguring = 0;
				clear_error = 0;
				error = 0;
			}
		}
		
		delay_write(writer_rc,writer_addr,writer_offset,writer_pattern,writer_v_pattern,
			&out_wrc,&out_waddr,&out_offset,&out_wpat,&out_wpatv);
		
		fprintf(fd,"%c ",'0' + configured);
		fprintf(fd,"%c ",'0' + source_cf);
		fprintf(fd,"%c ",'0' + drain_cf);
		fprintf(fd,"%c ",'0' + clear_error);
		fprintf(fd,"%c ",'0' + write_allow);
		
		write_flags(fd,flags);
		write_address_t(fd,src_base_address);
		write_address_t(fd,rcm_read_address);
		write_header_t(fd,read_header);
		write_address_t(fd,dest_base_address);
		write_address_t(fd,rcm_write_address);
		write_header_t(fd,write_header);
		
		fprintf(fd,"%c ",'0' + reader_rc);
		
		write_address_t(fd,reader_addr);
		write_control_t(fd,reader_pattern);
		write_control_t(fd,reader_v_pattern);
		write_address_t(fd,reader_offset);
		
		fprintf(fd,"%c ",'0' + out_wrc);
		
		write_address_t(fd,out_waddr);
		write_control_t(fd,out_wpat);
		write_control_t(fd,out_wpatv);
		write_address_t(fd,out_offset);
		
		fprintf(fd,"\n");
	}
	fclose(fd);

	return 0;
}


static void write_control_t(FILE* fd, control_t value) {
	//int i = OP_LENGTH-1;
	int i = (sizeof(control_t)*8)-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static void write_address_t(FILE* fd, address_t value) {
	//int i = OP_LENGTH-1;
	int i = (sizeof(address_t)*8)-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static void write_header_t(FILE* fd, header_t value) {
	//int i = (sizeof(header_t)*8)-1;
	int i = HEADER_SIZE-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static void write_flags(FILE* fd, char value) {
	int i = FLAG_SIZE-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static int bits_set(control_t value) {
	int i = (sizeof(control_t)*8)-1;
	int bits = 0;
	for (; i >= 0 ; i-- ) {
		if ( (value>>i)&0x01 == 1) {
			bits++;
		}
	}
	return bits;
}

static void generate_header(header_t* target,control_t* pattern, control_t* pattern_v, char* out_flags) {
	header_t new = 0;
	header_t version = 0x00;
	header_t flags = rand()&0x0F;
	header_t x_size = rand()%(CELLS_X+1);
	header_t y_size = rand()%(CELLS_Y+1);
	
	/* Build header */
	new = (version<<20)|(flags<<16)|(y_size<<8)|(x_size);

	*pattern = get_pattern(x_size);
	/* Schew the random result towards more valid configurations */
	if ( *pattern == 0 ) {
		*pattern = get_pattern(x_size);
	}
	*pattern_v = get_pattern(y_size);
	if ( *pattern_v == 0 ) {
		*pattern_v = get_pattern(y_size);
	}
	*out_flags = flags;
	*target = new;
}

control_t patterns[] = {
	0x0000,
	0x0001,
	0x0011,
	0x0000,
	0x0055,
	0x0000,
	0x0000,
	0x0000,
	0x00FF
};

static control_t get_pattern(header_t size) {
	if ( size > 8 ) {
		return 0;
	} else {
		return patterns[size];
	}
}

static void delay_write(char wrc, address_t waddr, address_t offset, control_t wpat, control_t wpatv,
		char* out_wrc, address_t* out_waddr, address_t* out_offset, control_t* out_wpat, control_t* out_wpatv) {

	static char wrc_d1 = 0;
	static address_t waddr_d1 = 0;
	static address_t offset_d1 = 0;
	static control_t wpat_d1 = 0;
	static control_t wpatv_d1 = 0;
	static char wrc_d2 = 0;
	static address_t waddr_d2 = 0;
	static address_t offset_d2 = 0;
	static control_t wpat_d2 = 0;
	static control_t wpatv_d2 = 0;
	
	*out_wrc = wrc_d2;
	*out_waddr = waddr_d2;
	*out_offset = offset_d2;
	*out_wpat = wpat_d2;
	*out_wpatv = wpatv_d2;
	
	wrc_d2 = wrc_d1;
	waddr_d2 = waddr_d1;
	offset_d2 = offset_d1;
	wpat_d2 = wpat_d1;
	wpatv_d2 = wpatv_d1;
	
	wrc_d1 = wrc;
	waddr_d1 = waddr;
	offset_d1 = offset;
	wpat_d1 = wpat;
	wpatv_d1 = wpatv;
}