#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#define bus_t uint32_t
#define control_t uint8_t
#define CELLS 8
#define STRIPE_SIZE 1
/* RC RCO RPC RPN LA RA RD RO CO */
/* ? This should use bitvectors*/
#define PATTERN "%c %d %hhd %hhd %d %hhd %d"

static void write_control_t(FILE* fd, control_t value);
static void write_bus_t(FILE* fd, bus_t value);
static int bits_set(bus_t value);


int main(int argc, char** argv) {
	long int vectors = 0;
	FILE* fd = NULL;
	char reconfigure = 0;
	char rcs_left = 0;
	control_t r_pattern = 0, r_pattern_next = 0, rc_to_units = 0;
	control_t unit_offset = 0;
	control_t vertical_pattern = 0;
	bus_t load_address = 0, read_address, next_address = 0, read_data = 0, out_data;
	
	time_t timeseed = time(NULL);
	srand(timeseed);
	
	if ( argc != 3 ) {
		printf("Usage: %s vectors <output filname>\n",argv[0]);
		return -1;
	}
	
	errno = 0;
	vectors = strtol(argv[1],NULL,10);
	if ( errno == ERANGE || errno == EINVAL ) {
		fprintf(stderr,"Parse error: %s\n",strerror(errno));
		return -1;
	}
	
	if ( (fd = fopen(argv[2],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	for (;vectors > 0 ; vectors-- ) {
		
		if ( reconfigure == 0 ) {
			reconfigure = !(rand()%4);
			if ( reconfigure == 1 ) {
				rcs_left == CELLS;
				r_pattern = rand();
				vertical_pattern = r_pattern;
				load_address = rand();
				read_address = load_address;
				rc_to_units = 1;
				if ( (r_pattern & 0x01) == 0 ) {
					out_data = 0;
					read_data = 0;
					next_address = load_address;
					unit_offset = 0;
				} else {
					out_data = read_data;
					read_data = rand();
					/* The next units load address should be unmodified 
					 * if the current unit it clocking dummy data */
					unit_offset = bits_set(vertical_pattern);
					next_address = load_address + unit_offset;
				}
			} else {
				out_data = 0;
				read_data = 0;
			}
		} else {
			vertical_pattern = (vertical_pattern>>1);
			rc_to_units = rc_to_units << 1;
			if ( rc_to_units  == 0 ) {
				reconfigure = 0;
			} 
			if ( (r_pattern & 0x01) == 0 || (vertical_pattern & 0x01 ) == 0 ) {
				out_data = read_data;
				read_data = 0;
			} else {
				out_data = read_data;
				read_data = rand();
				read_address += STRIPE_SIZE;
			}
		}
		
		r_pattern_next = (r_pattern>>1)|(r_pattern<<CELLS-1);
		
		fprintf(fd,"%c ",reconfigure + '0');
		write_control_t(fd,r_pattern);
		write_control_t(fd,r_pattern_next);
		write_control_t(fd,rc_to_units);
		write_bus_t(fd,load_address);
		write_bus_t(fd,read_address);
		write_bus_t(fd,read_data);
		write_bus_t(fd,out_data);
		write_bus_t(fd,next_address);
		fprintf(fd,"%c ",(vertical_pattern&0x01) + '0');
		write_bus_t(fd,unit_offset);
		fprintf(fd,"\n");
	}
	fclose(fd);

	return 0;
}

static void write_control_t(FILE* fd, control_t value) {
	//int i = OP_LENGTH-1;
	int i = (sizeof(control_t)*8)-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static void write_bus_t(FILE* fd, bus_t value) {
	//int i = OP_LENGTH-1;
	int i = (sizeof(bus_t)*8)-1;
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
}

static int bits_set(bus_t value) {
	int i = (sizeof(bus_t)*8)-1;
	int bits = 0;
	for (; i >= 0 ; i-- ) {
		if ( (value>>i)&0x01 == 1) {
			bits++;
		}
	}
	return bits;
}
