library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_shift_register_testbench is
end entity;

architecture shift_register_test of auto_shift_register_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;

	signal write_i		: std_logic;
	signal shift_left_i	: std_logic;
	signal shift_right_i	: std_logic;
	signal carry_in_i	: std_logic;
	
	signal data_i		: std_logic_vector(size -1 downto 0);

	signal carry_out_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);

	signal running		: integer := 1;
begin
	uut : shift_register
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			write => write_i,
			shift_left => shift_left_i,
			shift_right => shift_right_i,
			carry_in => carry_in_i,
			carry_out => carry_out_o,
			data_in => data_i,
			data_out => output_o
		);
	
	clock : process is
	begin
		while running = 1 loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			reset <= '1';
		end procedure;
		
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-shift_register.txt";
			variable L: line;
			variable good: boolean;
			
			variable operand	:integer;
			variable result		:integer;
		
			variable shift_left	:std_logic;
			variable shift_right	:std_logic;
			variable carry_in	:std_logic;
			variable carry_out	:std_logic;
		
			procedure perform_simulation is 
		
			begin
				write_i <= '1';
				carry_in_i <= carry_in;
				shift_left_i <= shift_left;
				shift_right_i <= shift_right;
				data_i <= std_logic_vector(to_signed(operand,size));
				
				wait until rising_edge(clk);
				
				wait for 1 ns; -- Allow for propagation
				assert carry_out = carry_out_o 
					report "Carry out incorrect " 
					& std_logic'image(shift_left) 
					& std_logic'image(shift_right) 
					& std_logic'image(carry_in) 
					& " " 
					& integer'image(operand)
					& " "
					& std_logic'image(carry_out) 
					& " was "
					& std_logic'image(carry_out_o)
					& integer'image(result) 
					severity failure;
				assert std_logic_vector(to_signed(result,size)) = output_o 
					report "Output incorrect "
					& std_logic'image(shift_left) 
					& std_logic'image(shift_right) 
					& std_logic'image(carry_in) 
					& " " 
					& integer'image(operand) 
					& " "
					& std_logic'image(carry_out) 
					& " "
					& integer'image(result)
					& " was "
					& integer'image(to_integer(signed(output_o))) 
					severity failure;
				
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,shift_left);
				read(L,shift_right);
				read(L,carry_in);
				
				read(L,operand);
				read(L,carry_out);
				read(L,result);
			
				perform_simulation;
			end loop;
		end procedure;
	begin
		reset_uut;
		
		random_stimuli;
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;