library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_shifter_testbench is
end entity;

architecture shifter_test of auto_shifter_testbench is
	constant size		: integer := 32;

	signal direction_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal output_o		: std_logic_vector(size -1 downto 0);
begin
	uut : shifter
		generic map(
			width => size
		)
		port map(
			direction => direction_i,
			data => data_a_i,
			shift_amount => data_b_i,
			output => output_o
		);
	
	random_test : process is
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-shifter.txt";
			variable L: line;
			
			variable first_operand	:integer;
			variable second_operand	:integer;
			variable result		:integer;
		
			variable dir		:std_logic;
		
			procedure perform_simulation is 
		
			begin
				direction_i  <= dir;
				data_a_i  <= std_logic_vector(to_signed(first_operand,size));
				
				data_b_i  <= std_logic_vector(to_signed(second_operand,size));
				
				wait for 100 ns; -- shouldn't this be valid at the same time ?
				assert output_o = std_logic_vector(to_signed(result,size)) report "Incorrect result" severity failure;
			
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,dir);
				read(L,first_operand);
				read(L,second_operand);
				read(L,result);
			
				perform_simulation;
			end loop;
		end procedure;
	begin
		direction_i <= '0';
		random_stimuli;
		-- Hack
		wait;
	end process;

end architecture;