library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common_defs.all;

entity shifter_testbench is
end entity;

architecture manual_shifter_test of shifter_testbench is
	constant size		: integer := 32;

	signal direction_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal output_o		: std_logic_vector(size -1 downto 0);
begin
	uut : shifter
		generic map(
			width => size
		)
		port map(
			direction => direction_i,
			data => data_a_i,
			shift_amount => data_b_i,
			output => output_o
		);
	
	manual_test : process
		procedure do_test(
			constant da	:in std_logic_vector(size-1 downto 0);
			constant db	:in std_logic_vector(size-1 downto 0);
			constant dir	:in std_logic;
			constant output	:in std_logic_vector(size-1 downto 0)) 
		is
		begin
			direction_i <= dir;
			data_a_i <= da;
			data_b_i <= db;
			
			wait for 100 ns;
			assert output_o = output report "Test failed" severity failure;
		end procedure;
	
	begin
		-- Test basic passtrough
		do_test(X"00000000",X"00000000",'0',X"00000000");
		do_test(X"11111111",X"00000000",'0',X"11111111");
		do_test(X"00000000",X"00000000",'1',X"00000000");
		do_test(X"11111111",X"00000000",'1',X"11111111");
		do_test(X"AAAAAAAA",X"00000000",'1',X"AAAAAAAA");
		do_test(X"55555555",X"00000000",'0',X"55555555");
		do_test(X"AAAAAAAA",X"00000000",'0',X"AAAAAAAA");
		do_test(X"55555555",X"00000000",'1',X"55555555");
		
		-- Test simple shifting
		do_test(X"00000100",X"00000001",'1',X"00000200");
		do_test(X"00000100",X"00000001",'0',X"00000080");
		
		-- Some simple cornercase shifting
		do_test(X"00000001",X"00000001",'1',X"00000002");
		do_test(X"80000000",X"00000001",'0',X"40000000");
		
		-- Complex shifting
		do_test(X"01010101",X"00000002",'1',X"04040404");
		do_test(X"08080808",X"00000003",'0',X"01010101");
		
		-- Wide shift (clears output)
		do_test(X"32423424",X"00010000",'1',X"00000000");
		do_test(X"435AFBCE",X"10400000",'0',X"00000000");
		
		-- Marginally wide shift (32 in either direction should clear)
		do_test(X"80000001",X"00000020",'1',X"00000000");
		do_test(X"80000001",X"00000020",'0',X"00000000");
		
		-- Marginally subwide shift (1 bit should be left high)
		do_test(X"80000001",X"0000001F",'1',X"80000000");
		do_test(X"80000001",X"0000001F",'0',X"00000001");
		
		-- Shift right border test
		do_test(X"00000002",X"00000000",'0',X"00000002");
		do_test(X"00000002",X"00000001",'0',X"00000001");
		do_test(X"00000002",X"00000002",'0',X"00000000");
		
		-- Shift left border test
		do_test(X"40000000",X"00000000",'1',X"40000000");
		do_test(X"40000000",X"00000001",'1',X"80000000");
		do_test(X"40000000",X"00000002",'1',X"00000000");
		
		wait;
	end process;
end architecture;