#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "generate.h"

static int allocate_cell (struct cell** target );
static void generate_configuration(struct cell* target);
static int check_loop(struct cell* target);
static void generate_alu_conf(struct cell* target);
static void generate_md_conf(struct cell* target);
static void generate_outputs(struct cell* target);
static void generate_init(struct cell* target);
static int allocate_structure(struct cell*** target, 
			int size_x, int size_y );

int generate_structure(struct cell*** target, 
			int size_x, int size_y ) {
	int i = 0,j = 0;
	struct cell* new = NULL;
	
	if ( allocate_structure(target,size_x,size_y) ) {
		return 1;
	}
	
	for ( i = 0; i < size_y; i++ ) {
		for ( j = 0 ; j < size_x ; j++ ) {
			if ( generate_cell(&new,j,i) ) {
				return 1;
			}
			MAP(*target,size_y,i,j) = new;
		}
	}
	return 0;
}

static int allocate_structure(struct cell*** target, 
			int size_x, int size_y ) {
	if ( target == NULL ) {
		fprintf(stderr,"Internal pointer error\n");
		return 2;
	}
	
	*target = malloc(sizeof(struct cell*)*size_x*size_y);
	
	if ( *target == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 1;
	}
	
	return 0;
}

void free_structure(struct cell*** target,
		int size_x, int size_y) {
	int i = 0, j = 0;
	struct cell* cell_target = NULL;
	if ( target == NULL ) {
		return;
	}
	for ( i = 0; i < size_y; i++ ) {
		for ( j = 0 ; j < size_x ; j++ ) {
			cell_target = MAP(*target,size_y,i,j);
			free(cell_target);
		}
	}
	free(*target);
	*target = NULL;
}


int generate_cell(struct cell** target, int x, int y) {

	if ( allocate_cell(target) ) {
		return 1;
	}
	
	generate_configuration(*target);
	(*target)->x = x;
	(*target)->y = y;
	return 0;
}

static void generate_configuration(struct cell* target) {
	
	do {
		generate_alu_conf(target);
		generate_md_conf(target);
	} while ( check_loop(target));
	
	generate_outputs(target);
	generate_init(target);
}

static int check_loop(struct cell* target ) {
	int md_in_alu = (target->alu_a == alu_mulout 
		|| target->alu_b == alu_mulout);
	
	int alu_in_md = (target->md_a == md_aluout 
		|| target->md_b == md_aluout);
	
	return (md_in_alu && alu_in_md);
}


static void generate_alu_conf(struct cell* target) {
	target->alu_a = rand()%ALU_INPUT;
	target->alu_b = rand()%ALU_INPUT;
	target->alu_op = rand()%ALU_OPERATIONS;
}

static void generate_md_conf(struct cell* target) {
	target->md_a = rand()%MD_INPUT;
	target->md_b = rand()%MD_INPUT;
	target->md_op = rand()%MD_OPERATIONS;
}

static void generate_outputs(struct cell* target) {
	target->state = rand()%OUTPUT;
	target->east = rand()%OUTPUT;
	target->south = rand()%OUTPUT;
}

static void generate_init(struct cell* target ) {
	int i = rand()%2;
	if ( i == 0 ) {
		target->do_init = !!(i);
		target->init_value = 0;
	} else {
		target->do_init = !!(i);
		target->init_value = rand();
	}
}


static int allocate_cell(struct cell** target) {
	
	if ( target == NULL ) {
		fprintf(stderr, "Invalid input to allocate_cell\n");
		return 1;
	}
	
	*target = malloc(sizeof(struct cell));
	
	if ( *target == NULL ) {
		fprintf(stderr, 
			"Memory allocation failed in allocate_cell\n");
		return 2;
	}
	
	return 0;
}