#!/bin/bash

CONFIGS=10
XSIZE=$WRPU_CX_SIZE
YSIZE=$WRPU_CY_SIZE
IDLE=$((($WRPU_X_SIZE/WRPU_CX_SIZE)-1))

SWSIM_INPUT=""
BIT_FILES=""
SIMULATOR=../../../simulator/simulator

if [ -e "./build" -a ! -d "./build" ]; then
	echo "Unable to create build directory"
	exit -1
else 
	rm -rf ./build
	mkdir ./build
fi

for i in $(seq 1 $CONFIGS) ; do
	mkdir ./build/sim$i
	cp buildMakefile ./build/sim$i/Makefile
	cp compute_defs.h ./build/sim$i
	./wrpu $XSIZE $YSIZE ./simulator-header.c build/sim$i/sim.c build/sim$i/trans.cell
	if [ $? -ne 0 ]; then
		echo "Generation failed"
		exit -1
	fi	
	cd ./build/sim$i/; make; cd ../../
	if [ $? -ne 0 ]; then
		echo "Make failed"
		exit -1
	fi
	../../../assembler/translator 1 $XSIZE $YSIZE ./build/sim$i/trans.cell ./build/sim$i/trans.bit >/dev/null
	if [ $? -ne 0 ]; then
		echo "Translator failed"
		exit -1
	fi
	SWSIM_INPUT="$SWSIM_INPUT ./build/sim$i/sim.so"
	BIT_FILES="$BIT_FILES ./build/sim$i/trans.bit"
done

echo $SWSIM_INPUT
echo $BIT_FILES
$SIMULATOR $XSIZE $YSIZE ./stats.txt ./io-dump.txt $IDLE 1 $SWSIM_INPUT > /dev/null

if [ $? -ne 0 ]; then
	echo "Simulator failed"
	exit -1
fi

./isim-wrapper.sh $BIT_FILES
./modelsim-wrapper.sh $BIT_FILES

