#include <stdio.h>
#include "cell.h"
#include "generate.h"
#include "translator_backend.h"

const char* trans_alu_operation[] = {
	"+",
	"-",
	"slr",
	"sll",
	"or",
	"nor",
	"and",
	"xor"
};

const char* trans_alu_input[] = {
	"west",
	"north",
	"state",
	"mulout"
};

const char* trans_md_operation[] = {
	"*",
	"/"
};

const char* trans_md_input[] = {
	"west",
	"north",
	"state",
	"aluout"
};

const char* trans_output[] = {
	"north",
	"west",
	"aluout",
	"mulout",
	"state"
};


static void write_cell(struct cell* source, FILE* fd);

int write_translator_file(struct cell** source, 
		int size_x, int size_y, FILE* fd) {
	int i = 0, j = 0;
	struct cell* source_cell = NULL;
	
	if ( fd == NULL ) {
		return 1;
	}
	
	for ( i = 0 ; i < size_y; i++ ) {
		for ( j = 0 ; j < size_x ; j++ ) {
			source_cell = MAP(source,size_y,i,j);
			write_cell(source_cell,fd);
		}
	}
	return 0;
}


static void write_cell(struct cell* source, FILE* fd) {
	fprintf(fd,"cell[%d][%d] {\n",source->y,source->x);
	fprintf(fd,"\taluout = %s %s %s;\n", 
		trans_alu_input[source->alu_a],
		trans_alu_operation[source->alu_op],
		trans_alu_input[source->alu_b]
	);
	fprintf(fd,"\tmulout = %s %s %s;\n",
		trans_md_input[source->md_a],
		trans_md_operation[source->md_op],
		trans_md_input[source->md_b]
	);
	fprintf(fd,"\teast = %s;\n",
		trans_output[source->east]
	);
	fprintf(fd,"\tsouth = %s;\n",
		trans_output[source->south]
	);
	fprintf(fd,"\tstate = %s;\n",
		trans_output[source->state]
	);
	if ( source->do_init == 1 ) {
		fprintf(fd,"\tinit {\n");
		fprintf(fd,"\t\tstate = %d;\n",source->init_value);
		fprintf(fd,"\t}\n");
	}
	fprintf(fd,"}\n");
}