#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include "cell.h"
#include "generate.h"
#include "translator_backend.h"
#include "simulator_backend.h"
#define BUFFER_SIZE 512

int parse_size(int* size, char* arg);

int main(int argc, char** argv) {
	char buffer[BUFFER_SIZE] = {0,};
	int bytes = 0;
	struct cell** target = NULL;
	int x_size = -1, y_size = -1;
	FILE* translator_output = NULL;
	FILE* simulator_hdr = NULL;
	FILE* simulator_output = NULL;
	
	struct timespec ts;
	clock_gettime(CLOCK_MONOTONIC,&ts);
	
	srand(ts.tv_nsec);
	
	if ( argc != 6 ) {
		printf("Usage: %s x-size y-size simulator-header"
			" simulator-output translator-output\n",
			argv[0]
		);
		return -1;
	}
	
	if ( parse_size(&x_size,argv[1]) ) {
		return -1;
	}
	if ( parse_size(&y_size,argv[2]) ) {
		return -1;
	}
	
	
	if ( generate_structure(&target,x_size,y_size )) {
		fprintf(stderr,"Cell generation failed\n");
		return -1;
	}
	
	if ( (translator_output = fopen(argv[5],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	if ( write_translator_file(target,x_size,y_size,translator_output) ) {
		fprintf(stderr,"Translator file generation failed\n");
		return -1;
	}
	fclose(translator_output);
	
	if ( (simulator_hdr = fopen(argv[3],"r")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	if ( (simulator_output = fopen(argv[4],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	while ( (bytes = fread(buffer,sizeof(char),BUFFER_SIZE,simulator_hdr))
			!= 0
		) {
	
		fwrite(buffer,sizeof(char),bytes,simulator_output);
	}
	fclose(simulator_hdr);
	
	
	if ( write_simulator_file(target,x_size,y_size,simulator_output) ) {
		fprintf(stderr,"Simulator file generation failed\n");
		return -1;
	}
	fclose(simulator_output);
	free_structure(&target,x_size,y_size);
	return 0;
}

int parse_size(int* size, char* arg) {
	int result = 0;
	errno = 0;
	
	result = strtol(arg,NULL,10);
	if ( errno == ERANGE || errno == EINVAL ) {
		fprintf(stderr,"Parse error %s\n",strerror(errno));
		return -1;
	}
	*size = result;
	return 0;
}