#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define __USE_GNU
#include <dlfcn.h>

#include "library.h"
#include "memory.h"
#include "cell.h"
#include "compute_defs.h"

#define MEMORY_WORDS 2048

struct mem_pattern* MEMCFG_SYMBOL = NULL;
memsize_t MEMORY_SIZE_SYMBOL = 0;
cell_transfer* MEMORY_SYMBOL = NULL;
grid_size_t GRID_DIMENTIONS;
grid_size_t GRID_SIZE[];
int random_seed;

static int generate_memory_pattern(int x, int y);
static int generate_memory(int x, int y);


void real_init(void ) {
	fprintf(stderr,"Testing library loaded\n");
	fprintf(stderr,"Autogenerated stub\n");
}

void finalize(void) {
	fprintf(stderr,"Library unloaded\n");
}

void
__attribute__ ((constructor))
local_init(void) {
	real_init();
}

void
__attribute__ ((destructor))
local_finit(void) {
	finalize();
}

void INIT_SYMBOL(grid_size_t dims, grid_size_t* sizes) {
	int i = 0;
	printf("Init running\n");
	srand(random_seed);
	printf("Dims: %d\n", dims);
	for ( i = 0 ; i < dims ; i++ ) {
		printf("Dimsize: %d\t",sizes[i]);
		if ( GRID_SIZE[i] != sizes[i] ) {
			printf("Invalid dimension\n");
			GRID_SIZE[i] = -1;
			return;
		}
	}
	printf("\n");
	if ( dims != 2 ) {
		printf("Only 2 dimensions are supported\n");
		return;
	}
	
	if ( !generate_memory(sizes[0],sizes[1]) ) {
		printf("Memory generation failed\n");
		return;
	}
	if ( !generate_memory_pattern(sizes[0],sizes[1]) ) {
		printf("Memory pattern generation failed\n");
		return;
	}
	printf("Memory pattern generated\n");
}

static int generate_memory(int x, int y) {
	int i = 0;
	
	MEMORY_SYMBOL = malloc(MEMORY_WORDS*(x+y)*sizeof(cell_transfer));
	if ( MEMORY_SYMBOL == NULL ) {
		return 0;
	}
	
	for ( i = 0; i < MEMORY_WORDS*(x+y) ;i++) {
		if ( rand()%4 == 0 ) {
			MEMORY_SYMBOL[i] = rand()%32768;
		} else {
			MEMORY_SYMBOL[i] = rand()%15;
		}
	}
	MEMORY_SIZE_SYMBOL = MEMORY_WORDS*(x+y)*sizeof(cell_transfer);
	
	return 1;
}


static int generate_memory_pattern(int x, int y) {
	int i = 0, j = 0;
	int grid_sum = x+y;
	struct mem_pattern* target = NULL;
	struct mem_pattern src = {0x0,0x00,0x2,0x0};
	struct mem_pattern drain = {0x00,0x00,0x2,0x0};
	
	MEMCFG_SYMBOL = malloc(sizeof(struct mem_pattern)*2*grid_sum);


	if ( MEMCFG_SYMBOL == NULL ) {
		return 0;
	}

	memset(MEMCFG_SYMBOL,0,sizeof(struct mem_pattern)*2*grid_sum);
	target = MEMCFG_SYMBOL;

	for ( i = 0 ; i < 2 ; i++ ) {
		for ( j = 0 ; j < x ; j++ ) {
			src.address = ((i*(x+y))+j)*MEMORY_WORDS;
			src.address_limit = ((i*(x+y))+(j+1))*MEMORY_WORDS;
			*(target+j+(i*grid_sum)) = src;
		}
		for ( j = 0 ; j < y ; j++ ) {
			drain.address = ((i*(x+y))+(j+x))*MEMORY_WORDS;
			drain.address_limit = ((i*(x+y))+(j+x+1))*MEMORY_WORDS;
			*(target+j+x+(i*grid_sum)) = drain;
		}
	}
	return 1;
}

void md_operation(int op, cell_transfer op1, cell_transfer op2, cell_transfer* target) {
	switch ( op ) {
	case md_mul:
		*target = op1 * op2;
		break;
	case md_div:
		if ( op2 == 0 ) {
			/* Div by zero */
			if ( op1 < 0 ) {
				*target = 0x8000;
			} else {
				*target = 0x7FFF;
			}
		} else {
			*target = op1 / op2;
		}
		break;
	default:
		fprintf(stderr,"Err. unknown mul op\n");
		exit(-1);
	}
}

void alu_operation(int op, cell_transfer op1, cell_transfer op2, cell_transfer* target) {
	u_cell_transfer shift_limit = ~0;
	shift_limit = log2(shift_limit);

	switch( op) {
	case alu_add:
		*target = op1 + op2;
		break;
	case alu_sub:
		*target = op1 - op2;
		break;
	case alu_slr:
		if ( (u_cell_transfer)op2 <= shift_limit ) {
			*target = (u_cell_transfer)op1 >> (u_cell_transfer)op2;
		} else {
			*target = 0;
		}
		break;
	case alu_sll:
		if ( (u_cell_transfer)op2 <= shift_limit ) {
			*target = (u_cell_transfer)op1 << (u_cell_transfer)op2;
		} else {
			*target = 0;
		}
		break;
	case alu_or:
		*target = op1 | op2;
		break;
	case alu_nor:
		*target = ~(op1 | op2);
		break;
	case alu_and:
		*target = op1 & op2;
		break;
	case alu_xor:
		*target = op1 ^ op2;
		break;
	default:
		fprintf(stderr,"Huh.. unknown aluop\n");
		exit(-1);
	}
}


