library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;
use work.version2_defs.all;

entity cell is
	generic (
		width		: positive := 16;
		config_width	: positive := 2
	);
	port (
		config_clk	: in std_logic;
		data_clk	: in std_logic;
		compute_clk	: in std_logic;
		
		reset		: in std_logic;
		
		reconfigure	: in std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0);
		
		north_idle	: in std_logic;
		west_idle	: in std_logic;
		south_idle	: out std_logic;
		east_idle	: out std_logic;
		
		-- With room for valid signal
		north		: in std_logic_vector(width downto 0);
		west		: in std_logic_vector(width downto 0);
		
		south		: out std_logic_vector(width downto 0);
		east		: out std_logic_vector(width downto 0)
		
	);
end entity;

architecture version2 of cell is
	constant exe_inputs_log	: positive	:= 2;
	constant data_outs_log	: positive	:= 3;
	constant config_bits	: positive	:= config_size-(width+1);
	constant out_valid_log	: positive	:= 2;
	constant branch_log	: positive	:= 2;
	constant switch_size	: positive	:= 2;
	constant in_branch_log	: positive	:= 2;
	constant in_switch_size	: positive	:= 2;
	
	signal north_reg	: std_logic_vector(width downto 0);
	signal west_reg		: std_logic_vector(width downto 0);
	
	signal north_ds_input	: std_logic_vector((2**in_branch_log)-1 downto 0);
	signal north_ds_sel	: std_logic_vector(in_branch_log-1 downto 0);
	signal north_ds_output	: std_logic_vector(0 downto 0);
	signal norwes_input	: std_logic_vector((in_switch_size*(width+1)) -1 downto 0);
	signal norwes_output	: std_logic_vector(width downto 0);
	
	signal west_ds_input	: std_logic_vector((2**in_branch_log)-1 downto 0);
	signal west_ds_sel	: std_logic_vector(in_branch_log-1 downto 0);
	signal west_ds_output	: std_logic_vector(0 downto 0);
	signal wesnor_input	: std_logic_vector((in_switch_size*(width+1)) -1 downto 0);
	signal wesnor_output	: std_logic_vector(width downto 0);
	
	-- Rooms for data + valid
	signal main_alu_input_a	: std_logic_vector(width downto 0);
	signal main_alu_input_b	: std_logic_vector(width downto 0);
	signal main_alu_op	: std_logic_vector(alu_op_size -1 downto 0);
	signal main_alu_out	: std_logic_vector(width downto 0);
	signal main_alu_flags	: std_logic_vector(alu_flags_size -1 downto 0);
	
	signal alu_input_a_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);
	signal alu_input_b_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);

	signal alu_mux_inputs	: std_logic_vector((2**exe_inputs_log)*(width+1) -1 downto 0);
	signal mul_mux_inputs	: std_logic_vector((2**exe_inputs_log)*(width+1) -1 downto 0);

	signal mul_input_a_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);
	signal mul_input_b_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);

	signal muldiv_input_a	: std_logic_vector(width downto 0);
	signal muldiv_input_b	: std_logic_vector(width downto 0);
	signal muldiv_op	: std_logic_vector(0 downto 0);
	signal muldiv_output	: std_logic_vector(width downto 0);

	signal state_output	: std_logic_vector(width downto 0);
	signal state_input	: std_logic_vector(width downto 0);
	
	signal state_write_input: std_logic_vector(1 downto 0);
	signal state_write	: std_logic_vector(0 downto 0);
	signal state_write_sel	: std_logic_vector(0 downto 0);
	
	signal state_mux_output	: std_logic_vector(width downto 0);
	
	signal state_v_mux_sel	: std_logic_vector(out_valid_log -1 downto 0);
	signal state_v_mux_input: std_logic_vector((2**out_valid_log)-1 downto 0);
	signal state_v_mux_out	: std_logic_vector(0 downto 0);
	
	signal state_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	
	signal south_v_mux_sel	: std_logic_vector(out_valid_log -1 downto 0);
	signal south_v_mux_input: std_logic_vector((2**out_valid_log)-1 downto 0);
	signal south_v_mux_out	: std_logic_vector(0 downto 0);
	
	signal soteas_ds_input	: std_logic_vector((2**branch_log)-1 downto 0);
	signal soteas_ds_sel	: std_logic_vector(branch_log -1 downto 0);
	signal soteas_ds_output	: std_logic_vector(0 downto 0);
	
	signal soteas_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal soteas_sw_output	: std_logic_vector(data_outs_log -1 downto 0);
	
	signal eassot_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal eassot_sw_output	: std_logic_vector(data_outs_log -1 downto 0);
	
	signal eassta_ds_input	: std_logic_vector((2**branch_log)-1 downto 0);
	signal eassta_ds_sel	: std_logic_vector(branch_log -1 downto 0);
	signal eassta_ds_output	: std_logic_vector(0 downto 0);
	
	signal eassta_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal eassta_sw_output	: std_logic_vector(data_outs_log -1 downto 0);
	
	signal staeas_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal staeas_sw_output	: std_logic_vector(data_outs_log -1 downto 0);
	
	signal stasot_ds_input	: std_logic_vector((2**branch_log)-1 downto 0);
	signal stasot_ds_sel	: std_logic_vector(branch_log -1 downto 0);
	signal stasot_ds_output	: std_logic_vector(0 downto 0);
	
	signal stasot_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal stasot_sw_output	: std_logic_vector(data_outs_log-1 downto 0);
	
	signal sotsta_sw_input	: std_logic_vector((data_outs_log*switch_size)-1 downto 0);
	signal sotsta_sw_output	: std_logic_vector(data_outs_log-1 downto 0);
	
	
	signal south_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	signal south_mux_output : std_logic_vector(width downto 0);
	
	signal east_v_mux_sel	: std_logic_vector(out_valid_log -1 downto 0);
	signal east_v_mux_input	: std_logic_vector((2**out_valid_log)-1 downto 0);
	signal east_v_mux_out	: std_logic_vector(0 downto 0);
	
	signal east_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	signal east_mux_output	: std_logic_vector(width downto 0);
	
	signal data_outputs	: std_logic_vector((2**data_outs_log)*(width+1) -1 downto 0);
	
	signal cs_bus		: std_logic_vector(config_width -1 downto 0);
	signal rc_bus		: std_logic_vector(config_width -1 downto 0);

	signal config		: std_logic_vector(config_bits-1 downto 0);
	signal config_data_in	: std_logic_vector(config_bits-1 downto 0);

	signal restart		: std_logic;
	signal reg_reconfigure	: std_logic;
	signal reg_idle		: std_logic;
	signal write_enable	: std_logic;
begin
	restart <= reset and (not data_clk);
	
	south_idle <= reg_idle;
	east_idle <= reg_idle;
	
	alu_input_mux_a: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => alu_mux_inputs,
			sel => alu_input_a_sel,
			output => main_alu_input_a
		);
	
	alu_input_mux_b: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => alu_mux_inputs,
			sel => alu_input_b_sel,
			output => main_alu_input_b
		);
	
	
	main_alu: component alu
		generic map(
			width => width
		)
		port map(
			clk => compute_clk,
			reset => restart,
			data_a => main_alu_input_a(width-1 downto 0),
			data_b => main_alu_input_b(width-1 downto 0),
			valid_a => main_alu_input_a(width),
			valid_b => main_alu_input_b(width),
			alu_op => main_alu_op,
			output => main_alu_out(width-1 downto 0),
			valid_output => main_alu_out(width),
			flags => main_alu_flags
		);
	
	muldiv_input_mux_a: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => mul_mux_inputs,
			sel => mul_input_a_sel,
			output => muldiv_input_a
		);
	
	muldiv_input_mux_b: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => mul_mux_inputs,
			sel => mul_input_b_sel,
			output => muldiv_input_b
		);
	
	muldiv: component multiplydivider
		generic map(
			width => width
		)
		port map(
			clk => compute_clk,
			reset => restart,
			opcode => muldiv_op(0),
			valid_a => muldiv_input_a(width),
			valid_b => muldiv_input_b(width),
			data_a => muldiv_input_a(width-1 downto 0),
			data_b => muldiv_input_b(width-1 downto 0),
			valid_output => muldiv_output(width),
			output => muldiv_output(width-1 downto 0)
		);
	
	state_input_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => stasot_sw_output,
			output => state_mux_output
		);
	
	state_v_mux_input <= "0" & main_alu_flags & state_mux_output(state_mux_output'left);
	state_valid_mux: component mux
		generic map(
			path_width => 1,
			sel_width => out_valid_log
		)
		port map(
			input => state_v_mux_input,
			sel => state_v_mux_sel,
			output => state_v_mux_out
		);
	
	state_input(state_input'left) <= state_v_mux_out(0) and state_mux_output(state_mux_output'left);
	state_input(state_input'left -1 downto 0) <= state_mux_output(state_mux_output'left -1 downto 0);
	
	state_register: component sc_register
		generic map(
			width => (width+1),
			config_width => config_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => cs_bus,
			config_out => rc_bus,
			config_enable => reg_reconfigure,
			data_in => state_input,
			data_out => state_output,
			write => write_enable
		);
	write_enable <= state_write(0) and not reg_idle;
	
	state_write_input <= state_input(state_input'left) & "1";
	state_write_mux: component mux
		generic map(
			path_width => 1,
			sel_width => 1
		)
		port map(
			input => state_write_input,
			sel => state_write_sel,
			output => state_write
		);
	
	
	
	config_register: component sc_register
		generic map(
			width => config_bits,
			config_width => config_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => config_in,
			config_out => cs_bus,
			config_enable => reg_reconfigure,
			data_in => config_data_in,
			data_out => config,
			write => '0'
		);
	config_data_in <= (others => '0');
	
	
	south_output_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => sotsta_sw_output,
			output => south_mux_output
		);
	
	
	-- Branch begin
	soteas_ds_input <= "0" & main_alu_flags & "0";
	
	south_east_do_switch_mux: component mux
		generic map(
			path_width => 1,
			sel_width => branch_log
		)
		port map(
			input => soteas_ds_input,
			sel => soteas_ds_sel,
			output => soteas_ds_output
		);
	
	soteas_sw_input <= east_mux_sel & south_mux_sel;
	
	south_east_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => soteas_sw_input,
			sel => soteas_ds_output,
			output => soteas_sw_output
		);
	
	eassot_sw_input <= south_mux_sel & east_mux_sel;
	east_south_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => eassot_sw_input,
			sel => soteas_ds_output,
			output => eassot_sw_output
		);
	
	eassta_ds_input <= "0" & main_alu_flags & "0";
	east_state_do_switch_mux: component mux
		generic map(
			path_width => 1,
			sel_width => branch_log
		)
		port map(
			input => eassta_ds_input,
			sel => eassta_ds_sel,
			output => eassta_ds_output
		);
	
	eassta_sw_input <= state_mux_sel & eassot_sw_output;
	east_state_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => eassta_sw_input,
			sel => eassta_ds_output,
			output => eassta_sw_output
		);
	
	staeas_sw_input <= eassot_sw_output & state_mux_sel;
	state_east_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => staeas_sw_input,
			sel => eassta_ds_output,
			output => staeas_sw_output
		);
	
	stasot_ds_input <= "0" & main_alu_flags & "0";
	state_south_do_switch_mux: component mux
		generic map(
			path_width => 1,
			sel_width => branch_log
		)
		port map(
			input => stasot_ds_input,
			sel => stasot_ds_sel,
			output => stasot_ds_output
		);
	
	stasot_sw_input <= soteas_sw_output & staeas_sw_output;
	state_south_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => stasot_sw_input,
			sel => stasot_ds_output,
			output => stasot_sw_output
		);
	
	sotsta_sw_input <= staeas_sw_output & soteas_sw_output;
	south_state_switch_mux: component mux
		generic map(
			path_width => data_outs_log,
			sel_width => 1
		)
		port map(
			input => sotsta_sw_input,
			sel => stasot_ds_output,
			output => sotsta_sw_output
		);
	-- Branch end 
	
	south_v_mux_input <= "0" & main_alu_flags & south_mux_output(south_mux_output'left);
	south_valid_mux: component mux
		generic map(
			path_width => 1,
			sel_width => out_valid_log
		)
		port map(
			input => south_v_mux_input,
			sel => south_v_mux_sel,
			output => south_v_mux_out
		);
	south(south'left) <= south_v_mux_out(0) and south_mux_output(south_mux_output'left) and not (reg_reconfigure or reg_idle);
	south(south'left -1 downto 0) <= south_mux_output(south_mux_output'left -1 downto 0);
	
	east_output_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => eassta_sw_output,
			output => east_mux_output
		);
	east_v_mux_input <= "0" & main_alu_flags & east_mux_output(east_mux_output'left);
	east_valid_mux: component mux
		generic map(
			path_width => 1,
			sel_width => out_valid_log
		)
		port map(
			input => east_v_mux_input,
			sel => east_v_mux_sel,
			output => east_v_mux_out
		);
	east(east'left) <= east_v_mux_out(0) and east_mux_output(east_mux_output'left) and not (reg_reconfigure or reg_idle);
	east(east'left -1 downto 0) <= east_mux_output(east_mux_output'left -1 downto 0);
	
	
	north_ds_input(3) <= '0'; -- Reserved
	north_ds_input(2) <= west_reg(west_reg'left);
	north_ds_input(1) <= not north_reg(north_reg'left);
	north_ds_input(0) <= '0';
	north_do_switch_mux : component mux
		generic map(
			path_width => 1,
			sel_width => in_branch_log
		)
		port map(
			input => north_ds_input,
			sel => north_ds_sel,
			output => north_ds_output
		);
	
	norwes_input <= west_reg & north_reg;
	north_west_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => 1
		)
		port map(
			input => norwes_input,
			sel => north_ds_output,
			output => norwes_output
		);
	
	west_ds_input(3) <= '0'; -- Reserved
	west_ds_input(2) <= north_reg(north_reg'left);
	west_ds_input(1) <= not west_reg(west_reg'left);
	west_ds_input(0) <= '0';
	west_do_switch_mux : component mux
		generic map(
			path_width => 1,
			sel_width => in_branch_log
		)
		port map(
			input => west_ds_input,
			sel => west_ds_sel,
			output => west_ds_output
		);
	
	wesnor_input <= north_reg & west_reg;
	west_north_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => 1
		)
		port map(
			input => wesnor_input,
			sel => west_ds_output,
			output => wesnor_output
		);
	
	
	regs : process(reset, data_clk)
	begin
		if reset = '0' then
			north_reg <= (north_reg'range => '0');
			west_reg <= (west_reg'range => '0');
			reg_reconfigure <= '0';
			reg_idle <= '0';
		elsif rising_edge(data_clk) then 
			north_reg <= north;
			west_reg <= west;
			reg_reconfigure <= reconfigure;
			reg_idle <= north_idle or west_idle;
		end if;
	end process;
	
	alu_mux_inputs <= muldiv_output & state_output & norwes_output & wesnor_output;
	mul_mux_inputs <= main_alu_out & state_output & norwes_output & wesnor_output;
	data_outputs(data_outputs'left downto 5*(width+1)) <= (others => '0');
	data_outputs(5*(width+1)-1 downto 0 ) <=  muldiv_output & main_alu_out & state_output & norwes_output & wesnor_output;
	
	north_ds_sel <= config(37 downto 36);
	west_ds_sel <= config(35 downto 34);
	stasot_ds_sel <= config(33 downto 32);
	eassta_ds_sel <= config(31 downto 30);
	soteas_ds_sel <= config(29 downto 28);
	state_v_mux_sel <= config(27 downto 26);
	east_v_mux_sel <= config(25 downto 24);
	south_v_mux_sel <= config(23 downto 22);
	state_write_sel <= config(21 downto 21);
	alu_input_a_sel <= config(20 downto 19);
	alu_input_b_sel <= config(18 downto 17);
	main_alu_op <= config(16 downto 14);
	mul_input_a_sel <= config(13 downto 12);
	mul_input_b_sel <= config(11 downto 10);
	muldiv_op <= config(9 downto 9);
	state_mux_sel <= config(8 downto 6);
	east_mux_sel <= config(5 downto 3);
	south_mux_sel <= config(2 downto 0);
	
	config_out <= rc_bus when reg_reconfigure = '1' else config_in;
	
end architecture;